from abc import ABC
from typing import Any, Optional, Tuple, Union

import torch
import torch.nn as nn

from brevitas.core.function_wrapper.misc import Identity
from brevitas.inject import BaseInjector as Injector
from brevitas.proxy.runtime_quant import ActQuantProxyFromInjectorBase
from brevitas.quant_tensor import FloatQuantTensor
from brevitas.utils.quant_utils import _CachedIOFloat


class ActFloatQuantProxyFromInjectorBase(ActQuantProxyFromInjectorBase, ABC):

    def scale(self, force_eval=True):
        return self.retrieve_attribute('scale', force_eval)

    def zero_point(self, force_eval=True):
        return self.retrieve_attribute('zero_point', force_eval)

    def exponent_bit_width(self, force_eval=True):
        return self.retrieve_attribute('exponent_bit_width', force_eval)

    def mantissa_bit_width(self, force_eval=True):
        return self.retrieve_attribute('mantissa_bit_width', force_eval)

    def exponent_bias(self, force_eval=True):
        return self.retrieve_attribute('exponent_bias', force_eval)

    def is_saturating(self, force_eval=True):
        return self.retrieve_attribute('saturating', force_eval)

    def inf_values(self, force_eval=True):
        return self.retrieve_attribute('inf_values', force_eval)

    def nan_values(self, force_eval=True):
        return self.retrieve_attribute('nan_values', force_eval)

    @property
    def input_view_impl(self):
        if self.fused_activation_quant_proxy.tensor_quant is not None:
            return self.fused_activation_quant_proxy.tensor_quant.input_view_impl
        else:
            return Identity()

    @property
    def is_ocp(self):
        is_e4m3 = self.mantissa_bit_width() == 3 and self.exponent_bit_width() == 4
        is_ocp_e4m3 = is_e4m3 and self.inf_values() is None and self.nan_values() == (('111',))

        is_e5m2 = self.mantissa_bit_width() == 2 and self.exponent_bit_width() == 5
        is_ocp_e5m2 = is_e5m2 and self.inf_values() == (
            ('00',)) and self.nan_values() == ('01', '11', '10')

        return is_ocp_e4m3 or is_ocp_e5m2

    @property
    def is_fnuz(self):
        is_e4m3 = self.mantissa_bit_width() == 3 and self.exponent_bit_width() == 4
        is_fnuz_e4m3 = is_e4m3 and self.inf_values() is None and self.nan_values(
        ) is None and self.exponent_bias() == 8

        is_e5m2 = self.mantissa_bit_width() == 2 and self.exponent_bit_width() == 5
        is_fnuz_e5m2 = is_e5m2 and self.inf_values() is None and self.nan_values(
        ) is None and self.exponent_bias() == 16
        return is_fnuz_e4m3 or is_fnuz_e5m2


class ActFloatQuantProxyFromInjector(ActFloatQuantProxyFromInjectorBase):

    def __init__(self, quant_layer: nn.Module, quant_injector: Injector):
        super().__init__(quant_layer, quant_injector)
        self.cache_class = _CachedIOFloat

    def create_quant_tensor(
            self,
            qt_args: Union[torch.Tensor, Tuple[Any]],
            x: Optional[FloatQuantTensor] = None) -> FloatQuantTensor:
        if x is None:
            out = FloatQuantTensor(*qt_args, signed=self.is_signed, training=self.training)
        else:
            out = FloatQuantTensor(
                qt_args,
                x.scale,
                x.zero_point,
                x.mantissa_bit_width,
                x.exponent_bit_width,
                x.exponent_bias,
                x.saturating,
                x.inf_values,
                x.nan_values,
                x.signed,
                self.training)
        return out



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Union

from torch import Tensor
from torch.nn import Upsample
from torch.nn import UpsamplingBilinear2d
from torch.nn import UpsamplingNearest2d
from torch.nn.functional import interpolate

from brevitas.function import round_ste
from brevitas.quant_tensor import QuantTensor

from .mixin.base import QuantLayerMixin


class QuantUpsample(QuantLayerMixin, Upsample):

    def __init__(
            self,
            size=None,
            scale_factor=None,
            mode='nearest',
            align_corners=None,
            return_quant_tensor: bool = True):
        Upsample.__init__(
            self, size=size, scale_factor=scale_factor, mode=mode, align_corners=align_corners)
        QuantLayerMixin.__init__(self, return_quant_tensor)

    @property
    def channelwise_separable(self) -> bool:
        return True

    @property
    def requires_export_handler(self):
        return False

    def forward(self, input: Union[Tensor, QuantTensor]):
        x = self.unpack_input(input)
        if self.export_mode:
            out = self.export_handler(x.value)
            self._set_global_is_quant_layer(False)
            return out
        y_value = interpolate(x.value, self.size, self.scale_factor, self.mode, self.align_corners)
        if self.mode != 'nearest':
            # round interpolated values to scale
            assert isinstance(x, QuantTensor), 'Input scale factor required to interpolate correctly'
            y_value = round_ste(y_value / x.scale) * x.scale
        y = x.set(value=y_value)
        return self.pack_output(y)


class QuantUpsamplingBilinear2d(QuantLayerMixin, UpsamplingBilinear2d):

    def __init__(self, size=None, scale_factor=None, return_quant_tensor: bool = True):
        UpsamplingBilinear2d.__init__(self, size=size, scale_factor=scale_factor)
        QuantLayerMixin.__init__(self, return_quant_tensor)

    @property
    def channelwise_separable(self) -> bool:
        return True

    @property
    def requires_export_handler(self):
        return False

    def forward(self, input: Union[Tensor, QuantTensor]):
        x = self.unpack_input(input)
        if self.export_mode:
            out = self.export_handler(x.value)
            self._set_global_is_quant_layer(False)
            return out
        y_value = interpolate(x.value, self.size, self.scale_factor, self.mode, self.align_corners)
        # round interpolated values to scale
        assert isinstance(x, QuantTensor), 'Input scale factor required to interpolate correctly'
        y_value = round_ste(y_value / x.scale) * x.scale
        y = x.set(value=y_value)
        return self.pack_output(y)


class QuantUpsamplingNearest2d(QuantLayerMixin, UpsamplingNearest2d):

    def __init__(self, size=None, scale_factor=None, return_quant_tensor: bool = True):
        UpsamplingNearest2d.__init__(self, size=size, scale_factor=scale_factor)
        QuantLayerMixin.__init__(self, return_quant_tensor)

    @property
    def channelwise_separable(self) -> bool:
        return True

    @property
    def requires_export_handler(self):
        return False

    def forward(self, input: Union[Tensor, QuantTensor]):
        x = self.unpack_input(input)
        if self.export_mode:
            out = self.export_handler(x.value)
            self._set_global_is_quant_layer(False)
            return out
        y_value = interpolate(x.value, self.size, self.scale_factor, self.mode, self.align_corners)
        y = x.set(value=y_value)
        return self.pack_output(y)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause


def filter_kwargs(kwargs_prefix, kwargs: dict):
    return {k[len(kwargs_prefix):]: v for (k, v) in kwargs.items() if k.startswith(kwargs_prefix)}



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from .act import ActQuantSolver
from .bias import BiasQuantSolver
from .clamp import ClampQuantSolver
from .trunc import TruncQuantSolver
from .weight import WeightQuantSolver



# This code was taken and modified from the Hugging Face Diffusers repository under the following
# LICENSE:

# Copyright 2024 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Mapping, Optional

import diffusers
from diffusers.models.attention_processor import Attention
from diffusers.models.lora import LoRACompatibleLinear
import packaging
import packaging.version
import torch
import torch.nn.functional as F

from brevitas.graph.base import ModuleInstanceToModuleInstance
from brevitas.nn.equalized_layer import EqualizedModule
from brevitas.nn.quant_activation import QuantIdentity
from brevitas.nn.quant_scale_bias import ScaleBias
from brevitas.quant_tensor import _unpack_quant_tensor


class QuantizableAttention(Attention):

    def __init__(
            self,
            query_dim: int,
            cross_attention_dim: Optional[int] = None,
            heads: int = 8,
            dim_head: int = 64,
            dropout: float = 0.0,
            bias=False,
            upcast_attention: bool = False,
            upcast_softmax: bool = False,
            cross_attention_norm: Optional[str] = None,
            cross_attention_norm_num_groups: int = 32,
            added_kv_proj_dim: Optional[int] = None,
            norm_num_groups: Optional[int] = None,
            spatial_norm_dim: Optional[int] = None,
            out_bias: bool = True,
            scale_qk: bool = True,
            only_cross_attention: bool = False,
            eps: float = 1e-5,
            rescale_output_factor: float = 1.0,
            residual_connection: bool = False,
            _from_deprecated_attn_block=False,
            dtype=torch.float32,
            processor: Optional["AttnProcessor"] = None):

        super().__init__(
            query_dim,
            cross_attention_dim,
            heads,
            dim_head,
            dropout,
            bias,
            upcast_attention,
            upcast_softmax,
            cross_attention_norm,
            cross_attention_norm_num_groups,
            added_kv_proj_dim,
            norm_num_groups,
            spatial_norm_dim,
            out_bias,
            scale_qk,
            only_cross_attention,
            eps,
            rescale_output_factor,
            residual_connection,
            _from_deprecated_attn_block,
            processor,
        )
        if self.to_q.weight.shape == self.to_k.weight.shape:
            self.to_qkv = LoRACompatibleLinear(
                query_dim, 3 * self.inner_dim, bias=bias, dtype=dtype)

            del self.to_q
            del self.to_k
            del self.to_v

        else:
            self.to_q = LoRACompatibleLinear(query_dim, self.inner_dim, bias=bias, dtype=dtype)
            self.to_kv = LoRACompatibleLinear(
                self.cross_attention_dim, 2 * self.inner_dim, bias=bias, dtype=dtype)

            del self.to_k
            del self.to_v

        self.to_out = torch.nn.ModuleList([])
        self.to_out.append(
            LoRACompatibleLinear(self.inner_dim, query_dim, bias=out_bias, dtype=dtype))
        self.to_out.append(torch.nn.Dropout(dropout))

    def load_state_dict(
            self, state_dict: Mapping[str, Any], strict: bool = True, assign: bool = False):
        if hasattr(self, 'to_qkv') and 'to_q.weight' in state_dict:
            new_weights = torch.cat(
                [state_dict['to_q.weight'], state_dict['to_k.weight'], state_dict['to_v.weight']],
                dim=0)
            state_dict['to_qkv.weight'] = new_weights

            del state_dict['to_q.weight']
            del state_dict['to_k.weight']
            del state_dict['to_v.weight']
        elif hasattr(self, 'to_kv') and 'to_k.weight' in state_dict:
            new_weights = torch.cat([state_dict['to_k.weight'], state_dict['to_v.weight']], dim=0)
            state_dict['to_kv.weight'] = new_weights
            del state_dict['to_k.weight']
            del state_dict['to_v.weight']
        return super().load_state_dict(state_dict, strict, assign)


class QuantAttentionLast(Attention):

    def __init__(
            self,
            query_dim: int,
            cross_attention_dim: Optional[int] = None,
            heads: int = 8,
            kv_heads: Optional[int] = None,
            dim_head: int = 64,
            dropout: float = 0.0,
            bias: bool = False,
            upcast_attention: bool = False,
            upcast_softmax: bool = False,
            cross_attention_norm: Optional[str] = None,
            cross_attention_norm_num_groups: int = 32,
            qk_norm: Optional[str] = None,
            added_kv_proj_dim: Optional[int] = None,
            added_proj_bias: Optional[bool] = True,
            norm_num_groups: Optional[int] = None,
            spatial_norm_dim: Optional[int] = None,
            out_bias: bool = True,
            scale_qk: bool = True,
            only_cross_attention: bool = False,
            eps: float = 1e-5,
            rescale_output_factor: float = 1.0,
            residual_connection: bool = False,
            _from_deprecated_attn_block: bool = False,
            processor: Optional["AttnProcessor"] = None,
            out_dim: int = None,
            context_pre_only=None,
            pre_only=False,
            matmul_input_quant=None,
            is_equalized=False,
            fuse_qkv=False):

        super().__init__(
            query_dim,
            cross_attention_dim,
            heads,
            kv_heads,
            dim_head,
            dropout,
            bias,
            upcast_attention,
            upcast_softmax,
            cross_attention_norm,
            cross_attention_norm_num_groups,
            qk_norm,
            added_kv_proj_dim,
            added_proj_bias,
            norm_num_groups,
            spatial_norm_dim,
            out_bias,
            scale_qk,
            only_cross_attention,
            eps,
            rescale_output_factor,
            residual_connection,
            _from_deprecated_attn_block,
            processor,
            out_dim,
            context_pre_only,
            pre_only,
        )
        if fuse_qkv:
            self.fuse_projections()

        self.output_softmax_quant = QuantIdentity(matmul_input_quant)
        self.out_q = QuantIdentity(matmul_input_quant)
        self.out_k = QuantIdentity(matmul_input_quant)
        self.out_v = QuantIdentity(matmul_input_quant)
        if is_equalized:
            replacements = []
            for n, m in self.named_modules():
                if isinstance(m, torch.nn.Linear):
                    in_channels = m.in_features
                    eq_m = EqualizedModule(ScaleBias(in_channels, False, (1, 1, -1)), m)
                    r = ModuleInstanceToModuleInstance(m, eq_m)
                    replacements.append(r)
            for r in replacements:
                r.apply(self)

    def get_attention_scores(
            self,
            query: torch.Tensor,
            key: torch.Tensor,
            attention_mask: torch.Tensor = None) -> torch.Tensor:
        r"""
        Compute the attention scores.

        Args:
            query (`torch.Tensor`): The query tensor.
            key (`torch.Tensor`): The key tensor.
            attention_mask (`torch.Tensor`, *optional*): The attention mask to use. If `None`, no mask is applied.

        Returns:
            `torch.Tensor`: The attention probabilities/scores.
        """
        dtype = query.dtype
        if self.upcast_attention:
            query = query.float()
            key = key.float()

        if attention_mask is None:
            baddbmm_input = torch.empty(
                query.shape[0],
                query.shape[1],
                key.shape[1],
                dtype=query.dtype,
                device=query.device)
            beta = 0
        else:
            baddbmm_input = attention_mask
            beta = 1

        attention_scores = torch.baddbmm(
            baddbmm_input,
            query,
            key.transpose(-1, -2),
            beta=beta,
            alpha=self.scale,
        )
        del baddbmm_input

        if self.upcast_softmax:
            attention_scores = attention_scores.float()

        attention_probs = attention_scores.softmax(dim=-1)
        del attention_scores

        attention_probs = attention_probs.to(dtype)

        attention_probs = _unpack_quant_tensor(self.output_softmax_quant(attention_probs))
        return attention_probs


class QuantAttention(QuantizableAttention):

    def __init__(
            self,
            query_dim: int,
            cross_attention_dim: Optional[int] = None,
            heads: int = 8,
            dim_head: int = 64,
            dropout: float = 0.0,
            bias=False,
            upcast_attention: bool = False,
            upcast_softmax: bool = False,
            cross_attention_norm: Optional[str] = None,
            cross_attention_norm_num_groups: int = 32,
            added_kv_proj_dim: Optional[int] = None,
            norm_num_groups: Optional[int] = None,
            spatial_norm_dim: Optional[int] = None,
            out_bias: bool = True,
            scale_qk: bool = True,
            only_cross_attention: bool = False,
            eps: float = 1e-5,
            rescale_output_factor: float = 1.0,
            residual_connection: bool = False,
            _from_deprecated_attn_block=False,
            processor: Optional["AttnProcessor"] = None,
            matmul_input_quant=None,
            dtype=torch.float32,
            is_equalized=False):

        super().__init__(
            query_dim,
            cross_attention_dim,
            heads,
            dim_head,
            dropout,
            bias,
            upcast_attention,
            upcast_softmax,
            cross_attention_norm,
            cross_attention_norm_num_groups,
            added_kv_proj_dim,
            norm_num_groups,
            spatial_norm_dim,
            out_bias,
            scale_qk,
            only_cross_attention,
            eps,
            rescale_output_factor,
            residual_connection,
            _from_deprecated_attn_block,
            dtype,
            processor,
        )
        self.output_softmax_quant = QuantIdentity(matmul_input_quant)
        self.out_q = QuantIdentity(matmul_input_quant)
        self.out_k = QuantIdentity(matmul_input_quant)
        self.out_v = QuantIdentity(matmul_input_quant)
        if is_equalized:
            replacements = []
            for n, m in self.named_modules():
                if isinstance(m, torch.nn.Linear):
                    in_channels = m.in_features
                    eq_m = EqualizedModule(ScaleBias(in_channels, False, (1, 1, -1)), m)
                    r = ModuleInstanceToModuleInstance(m, eq_m)
                    replacements.append(r)
            for r in replacements:
                r.apply(self)

    def get_attention_scores(
            self,
            query: torch.Tensor,
            key: torch.Tensor,
            attention_mask: torch.Tensor = None) -> torch.Tensor:
        r"""
        Compute the attention scores.

        Args:
            query (`torch.Tensor`): The query tensor.
            key (`torch.Tensor`): The key tensor.
            attention_mask (`torch.Tensor`, *optional*): The attention mask to use. If `None`, no mask is applied.

        Returns:
            `torch.Tensor`: The attention probabilities/scores.
        """
        dtype = query.dtype
        if self.upcast_attention:
            query = query.float()
            key = key.float()

        if attention_mask is None:
            baddbmm_input = torch.empty(
                query.shape[0],
                query.shape[1],
                key.shape[1],
                dtype=query.dtype,
                device=query.device)
            beta = 0
        else:
            baddbmm_input = attention_mask
            beta = 1

        attention_scores = torch.baddbmm(
            baddbmm_input,
            query,
            key.transpose(-1, -2),
            beta=beta,
            alpha=self.scale,
        )
        del baddbmm_input

        if self.upcast_softmax:
            attention_scores = attention_scores.float()

        attention_probs = attention_scores.softmax(dim=-1)
        del attention_scores

        attention_probs = attention_probs.to(dtype)

        attention_probs = _unpack_quant_tensor(self.output_softmax_quant(attention_probs))
        return attention_probs


class AttnProcessor:
    r"""
    Default processor for performing attention-related computations.
    """

    def __call__(
        self,
        attn: Attention,
        hidden_states,
        encoder_hidden_states=None,
        attention_mask=None,
        temb=None,
        scale=1.0,
    ):
        residual = hidden_states

        if attn.spatial_norm is not None:
            hidden_states = attn.spatial_norm(hidden_states, temb)

        input_ndim = hidden_states.ndim

        if input_ndim == 4:
            batch_size, channel, height, width = hidden_states.shape
            hidden_states = hidden_states.view(batch_size, channel, height * width).transpose(1, 2)

        batch_size, sequence_length, _ = (
            hidden_states.shape if encoder_hidden_states is None else encoder_hidden_states.shape
        )
        attention_mask = attn.prepare_attention_mask(attention_mask, sequence_length, batch_size)

        if attn.group_norm is not None:
            hidden_states = attn.group_norm(hidden_states.transpose(1, 2)).transpose(1, 2)

        if encoder_hidden_states is None:
            assert attn.norm_cross is None, "Not supported"
            query, key, value = attn.to_qkv(hidden_states, scale=scale).chunk(3, dim=-1)

        else:
            assert not hasattr(attn, 'to_qkv'), 'Model not created correctly'
            query = attn.to_q(hidden_states, scale=scale)
            if attn.norm_cross:
                encoder_hidden_states = attn.norm_encoder_hidden_states(encoder_hidden_states)
            key, value = attn.to_kv(encoder_hidden_states, scale=scale).chunk(2, dim=-1)
        if hasattr(attn, 'out_q'):
            query = _unpack_quant_tensor(attn.out_q(query))
        if hasattr(attn, 'out_k'):
            key = _unpack_quant_tensor(attn.out_k(key))
        if hasattr(attn, 'out_v'):
            value = _unpack_quant_tensor(attn.out_v(value))
        query = attn.head_to_batch_dim(query)
        key = attn.head_to_batch_dim(key)
        value = attn.head_to_batch_dim(value)

        attention_probs = attn.get_attention_scores(query, key, attention_mask)
        hidden_states = torch.bmm(attention_probs, value)
        hidden_states = attn.batch_to_head_dim(hidden_states)

        # linear proj
        hidden_states = attn.to_out[0](hidden_states, scale=scale)
        # dropout
        hidden_states = attn.to_out[1](hidden_states)

        if input_ndim == 4:
            hidden_states = hidden_states.transpose(-1,
                                                    -2).reshape(batch_size, channel, height, width)

        if attn.residual_connection:
            hidden_states = hidden_states + residual

        hidden_states = hidden_states / attn.rescale_output_factor

        return hidden_states


class AttnProcessor2_0:
    r"""
    Processor for implementing scaled dot-product attention (enabled by default if you're using PyTorch 2.0).
    """

    def __init__(self):
        if not hasattr(F, "scaled_dot_product_attention"):
            raise ImportError(
                "AttnProcessor2_0 requires PyTorch 2.0, to use it, please upgrade PyTorch to 2.0.")

    def __call__(
        self,
        attn: Attention,
        hidden_states,
        encoder_hidden_states=None,
        attention_mask=None,
        temb=None,
        scale: float = 1.0,
    ):
        residual = hidden_states

        if attn.spatial_norm is not None:
            hidden_states = attn.spatial_norm(hidden_states, temb)

        input_ndim = hidden_states.ndim

        if input_ndim == 4:
            batch_size, channel, height, width = hidden_states.shape
            hidden_states = hidden_states.view(batch_size, channel, height * width).transpose(1, 2)

        batch_size, sequence_length, _ = (
            hidden_states.shape if encoder_hidden_states is None else encoder_hidden_states.shape
        )

        if attention_mask is not None:
            attention_mask = attn.prepare_attention_mask(
                attention_mask, sequence_length, batch_size)
            # scaled_dot_product_attention expects attention_mask shape to be
            # (batch, heads, source_length, target_length)
            attention_mask = attention_mask.view(
                batch_size, attn.heads, -1, attention_mask.shape[-1])

        if attn.group_norm is not None:
            hidden_states = attn.group_norm(hidden_states.transpose(1, 2)).transpose(1, 2)

        if encoder_hidden_states is None:
            assert attn.norm_cross is None, "Not supported"
            query, key, value = attn.to_qkv(hidden_states, scale=scale).chunk(3, dim=-1)

        else:
            assert not hasattr(attn, 'to_qkv'), 'Model not created correctly'
            query = attn.to_q(hidden_states, scale=scale)
            if attn.norm_cross:
                encoder_hidden_states = attn.norm_encoder_hidden_states(encoder_hidden_states)
            key, value = attn.to_kv(encoder_hidden_states, scale=scale).chunk(2, dim=-1)

        inner_dim = key.shape[-1]
        head_dim = inner_dim // attn.heads

        query = query.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)

        key = key.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)
        value = value.view(batch_size, -1, attn.heads, head_dim).transpose(1, 2)

        # the output of sdp = (batch, num_heads, seq_len, head_dim)
        # TODO: add support for attn.scale when we move to Torch 2.1
        hidden_states = F.scaled_dot_product_attention(
            query, key, value, attn_mask=attention_mask, dropout_p=0.0, is_causal=False)

        hidden_states = hidden_states.transpose(1, 2).reshape(batch_size, -1, attn.heads * head_dim)
        hidden_states = hidden_states.to(query.dtype)

        # linear proj
        hidden_states = attn.to_out[0](hidden_states, scale=scale)
        # dropout
        hidden_states = attn.to_out[1](hidden_states)

        if input_ndim == 4:
            hidden_states = hidden_states.transpose(-1,
                                                    -2).reshape(batch_size, channel, height, width)

        if attn.residual_connection:
            hidden_states = hidden_states + residual

        hidden_states = hidden_states / attn.rescale_output_factor

        return hidden_states



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from .base_quant_tensor import *
from .base_quant_tensor import _unpack_quant_tensor
from .float_quant_tensor import *
from .groupwise_float_quant_tensor import *
from .groupwise_int_quant_tensor import *
from .int_quant_tensor import *



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import List, Optional, Tuple

import torch
from torch import nn
from torch import Tensor

import brevitas
import brevitas.config as config
from brevitas.core.utils import inplace_momentum_update
from brevitas.core.utils import inplace_tensor_mul

from .view_wrapper import _ViewCatParameterWrapper
from .view_wrapper import _ViewParameterWrapper

DEFAULT_MOMENTUM = 0.1
SCALAR_SHAPE = ()


class _Stats(brevitas.jit.ScriptModule):
    __constants__ = ['stats_output_shape']

    def __init__(self, stats_impl: nn.Module, stats_output_shape: Tuple[int, ...]) -> None:
        super(_Stats, self).__init__()
        self.stats_output_shape = stats_output_shape
        self.stats_impl = stats_impl

    @brevitas.jit.script_method
    def forward(self, input: Tensor) -> Tensor:
        stats = self.stats_impl(input)
        stats = stats.view(self.stats_output_shape)
        return stats


class _RuntimeStats(brevitas.jit.ScriptModule):
    __constants__ = ['stats_input_concat_dim', 'stats_permute_dims', 'momentum']

    def __init__(
            self,
            stats_impl: nn.Module,
            stats_output_shape: Tuple[int, ...],
            stats_input_view_shape_impl: nn.Module,
            stats_buffer_momentum: float,
            dtype: Optional[torch.dtype],
            device: Optional[torch.device]) -> None:
        super(_RuntimeStats, self).__init__()
        self.counter = brevitas.jit.Attribute(0, int)
        self.stats_input_view_shape_impl = stats_input_view_shape_impl
        self.stats = _Stats(stats_impl, stats_output_shape)
        self.momentum = stats_buffer_momentum
        self.register_buffer(
            'running_stats', torch.full(stats_output_shape, 1.0, dtype=dtype, device=device))

    @brevitas.jit.script_method
    def forward(self, stats_input) -> Tensor:
        if self.training:
            stats_input = self.stats_input_view_shape_impl(stats_input)
            stats = self.stats(stats_input)
            new_counter = self.counter + 1
            if self.counter == 0:
                inplace_tensor_mul(self.running_stats, stats.detach())
            else:
                inplace_momentum_update(
                    self.running_stats, stats.detach(), self.momentum, self.counter, new_counter)
            self.counter = new_counter
        else:
            stats = self.running_stats
        return stats

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        super(_RuntimeStats, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        running_stats_key = prefix + 'running_stats'
        if config.IGNORE_MISSING_KEYS and running_stats_key in missing_keys:
            missing_keys.remove(running_stats_key)
        # Pytorch stores training flag as a buffer with JIT enabled
        training_key = prefix + 'training'
        if training_key in missing_keys:
            missing_keys.remove(training_key)


class _ParameterListStats(brevitas.jit.ScriptModule):
    __constants__ = ['stats_input_concat_dim']

    def __init__(
            self,
            stats_impl: nn.Module,
            stats_output_shape: Tuple[int, ...],
            stats_input_view_shape_impl: nn.Module,
            stats_input_concat_dim: int,
            tracked_parameter_list: List[torch.nn.Parameter]) -> None:
        super(_ParameterListStats, self).__init__()

        self.stats_input_concat_dim = stats_input_concat_dim
        self.first_tracked_param = _ViewParameterWrapper(
            tracked_parameter_list[0], stats_input_view_shape_impl)
        if len(tracked_parameter_list) > 1:
            extra_list = [
                _ViewCatParameterWrapper(
                    param, stats_input_view_shape_impl, stats_input_concat_dim)
                for param in tracked_parameter_list[1:]]
            self.extra_tracked_params_list = torch.nn.ModuleList(extra_list)
        else:
            self.extra_tracked_params_list = None
        self.stats = _Stats(stats_impl, stats_output_shape)

    @brevitas.jit.script_method
    def forward(self) -> torch.Tensor:
        stats_input = self.first_tracked_param()
        if self.extra_tracked_params_list is not None:
            for extra_tracked_param in self.extra_tracked_params_list:
                stats_input = extra_tracked_param(stats_input)
        out = self.stats(stats_input)
        return out



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import List, Optional, Tuple

import torch
from torch import Tensor

import brevitas
from brevitas.core.utils import StatelessBuffer
from brevitas.function.ops import max_float


class FloatScaling(brevitas.jit.ScriptModule):

    def __init__(
            self,
            max_available_float: Optional[float] = None,
            inf_values: Optional[Tuple[str]] = None,
            nan_values: Optional[Tuple[str]] = None,
            saturating: bool = True,
            device: Optional[str] = None,
            dtype: Optional[torch.dtype] = None):
        super(FloatScaling, self).__init__()
        self.inf_values = inf_values
        self.nan_values = nan_values
        self.saturating = saturating

        if max_available_float:
            max_available_float = torch.tensor(max_available_float, device=device, dtype=dtype)
            self.max_available_float = StatelessBuffer(max_available_float)
        else:
            self.max_available_float = None

    @brevitas.jit.script_method
    def forward(
            self, exponent_bit_width: Tensor, mantissa_bit_width: Tensor,
            exponent_bias: Tensor) -> Tensor:
        max_value = max_float(exponent_bit_width, mantissa_bit_width, exponent_bias)
        max_value = max_value if self.max_available_float is None else torch.min(
            max_value, self.max_available_float())
        return max_value



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
SPDX-License-Identifier: MIT
"""

import argparse
import re


class CustomValidator(object):

    def __init__(self, pattern):
        self._pattern = re.compile(pattern)

    def __call__(self, value):
        if not self._pattern.findall(value):
            raise argparse.ArgumentTypeError(
                "Argument has to match '{}'".format(self._pattern.pattern))
        return value


quant_format_validator = CustomValidator(r"int|e[1-8]m[1-8]")


def add_bool_arg(parser, name, default, help, str_true=False):
    dest = name.replace('-', '_')
    group = parser.add_mutually_exclusive_group(required=False)
    if str_true:
        group.add_argument('--' + name, dest=dest, type=str, help=help)
    else:
        group.add_argument('--' + name, dest=dest, action='store_true', help='Enable ' + help)
    group.add_argument('--no-' + name, dest=dest, action='store_false', help='Disable ' + help)
    parser.set_defaults(**{dest: default})



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from argparse import Namespace
import warnings

import numpy as np
from torch import Tensor
import torch.nn as nn
from torch.utils.data import DataLoader

from brevitas.export import export_onnx_qcdq
from brevitas.export import export_qonnx
from brevitas.export import export_torch_qcdq

from .train import device


def save_model_io(inp: Tensor, out: Tensor, save_path: str):
    with open(f"{save_path}/input.npy", "wb") as f:
        np.save(f, inp.cpu().numpy())
    with open(f"{save_path}/output.npy", "wb") as f:
        np.save(f, out.cpu().numpy())


def export(model: nn.Module, testloader: DataLoader, args: Namespace, opset_version: int = 11):
    save_path = args.save_path
    inp = testloader.dataset[0][0].unsqueeze(0)  # NCHW
    if args.save_model_io:
        inp = inp.to(device)
        model = model.to(device)
        save_model_io(inp, model(inp).detach(), save_path)
        print(f"Saved I/O to {save_path} as numpy arrays")
    if args.export_to_qonnx:
        export_qonnx(
            model.cpu(),
            input_t=inp.cpu(),
            export_path=f"{save_path}/qonnx_model.onnx",
            opset_version=opset_version)
        print(f"Saved QONNX model to {save_path}/qonnx_model.onnx")
    if args.export_to_qcdq_onnx:
        if opset_version < 13:
            warnings.warn("Need opset 13+ to support per-channel quantization.")
        else:
            export_onnx_qcdq(
                model.cpu(),
                input_t=inp.cpu(),
                export_path=f"{save_path}/qcdq_onnx_model.onnx",
                opset_version=opset_version)
            print(f"Saved QCDQ ONNX model to {save_path}/qcdq_onnx_model.onnx")
    if args.export_to_qcdq_torch:
        export_torch_qcdq(
            model.cpu(), input_t=inp.cpu(), export_path=f"{save_path}/qcdq_torch_model.pt")
        print(f"Saved QCDQ TorchScript model to {save_path}/qcdq_torch_model.pt")



import functools
import math
from typing import Callable
import warnings

import torch
from torch import Tensor
import torch.nn.functional as F

from brevitas.function.ops import max_int
from brevitas.function.ops_ste import ceil_ste
from brevitas.utils.torch_utils import compute_channel_view_shape

INT_QUANT_TENSOR_FN_HANDLER = {}


def implements_int_qt(torch_function):

    @functools.wraps(torch_function)
    def decorator(func):
        INT_QUANT_TENSOR_FN_HANDLER[torch_function] = func
        return func

    return decorator


@implements_int_qt(torch.cat)
def cat_handler(*args, **kwargs):
    from brevitas.quant_tensor import IntQuantTensor
    return IntQuantTensor.cat(*args, **kwargs)


@implements_int_qt(F.conv1d)
def conv1d_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.conv1d, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_int_qt(F.conv2d)
def conv2d_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.conv2d, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_int_qt(F.conv3d)
def conv3d_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.conv3d, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_int_qt(F.conv_transpose1d)
def conv_transpose1d_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.conv_transpose1d, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_int_qt(F.conv_transpose2d)
def conv_transpose2d_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.conv_transpose2d, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_int_qt(F.conv_transpose3d)
def conv_transpose3d_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.conv_transpose3d, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_int_qt(F.linear)
def linear_handler(quant_input, quant_weight, bias=None, *args, **kwargs):
    output = quant_layer(F.linear, quant_input, quant_weight, bias, *args, **kwargs)
    return output


@implements_int_qt(F.embedding)
def embedding_handler(input, quant_weight, *args, **kwargs):
    from brevitas.quant_tensor import _unpack_quant_tensor
    from brevitas.quant_tensor import IntQuantTensor

    quant_weight_value = _unpack_quant_tensor(quant_weight)
    out = F.embedding(input, quant_weight_value, *args, **kwargs)

    if isinstance(quant_weight, IntQuantTensor):
        scale = quant_weight.scale
        zero_point = quant_weight.zero_point
        bit_width = quant_weight.bit_width
        if any(t.numel() > 1 for t in [scale, zero_point, bit_width]):
            raise RuntimeError("Only per-tensor quantization is supported.")
        signed = quant_weight.signed
        training = quant_weight.training
        out = IntQuantTensor(out, scale, zero_point, bit_width, signed, training)
    return out


@implements_int_qt(F.avg_pool2d)
def avg_pool2d_handler(
        quant_input, kernel_size, stride, padding, ceil_mode, count_include_pad, divisor_override):
    from brevitas.quant_tensor import _unpack_quant_tensor

    x = F.avg_pool2d(
        _unpack_quant_tensor(quant_input),
        kernel_size,
        stride,
        padding,
        ceil_mode,
        count_include_pad,
        divisor_override)

    max_acc_bit_width = FN_ACC_BITWIDTH_MAPPING[F.avg_pool2d]
    # remove avg scaling
    if isinstance(kernel_size, tuple):
        avg_scaling = kernel_size[0] * kernel_size[1]
    else:
        avg_scaling = kernel_size * kernel_size

    quant_input = quant_input.set(value=x)
    quant_input = quant_input.set(bit_width=max_acc_bit_width(quant_input.bit_width, avg_scaling))
    return quant_input


@implements_int_qt(F.adaptive_avg_pool2d)
def adaptive_avg_pool2d_handler(quant_input, output_shape):
    from functools import reduce
    from operator import mul

    from brevitas.nn.quant_avg_pool import TruncAdaptiveAvgPool2d
    from brevitas.quant_tensor import _unpack_quant_tensor

    x = F.adaptive_avg_pool2d(_unpack_quant_tensor(quant_input), output_shape)
    k_size, stride = TruncAdaptiveAvgPool2d.compute_kernel_size_stride(quant_input.value.shape[2:], x.shape[2:])

    max_acc_bit_width = FN_ACC_BITWIDTH_MAPPING[F.avg_pool2d]
    reduce_size = reduce(mul, k_size, 1)

    quant_input = quant_input.set(value=x)
    quant_input = quant_input.set(bit_width=max_acc_bit_width(quant_input.bit_width, reduce_size))
    return quant_input


def quant_layer(fn, quant_input, quant_weight, bias, *args, **kwargs):
    from brevitas.quant_tensor import _unpack_quant_tensor
    from brevitas.quant_tensor import IntQuantTensor

    output_scale = None
    output_bit_width = None
    output_zero_point = None
    output_signed = None
    max_acc_bit_width = FN_ACC_BITWIDTH_MAPPING[fn]

    compute_output_quant_tensor = isinstance(quant_input, IntQuantTensor) and isinstance(
        quant_weight, IntQuantTensor)

    if bias is None:
        output = fn(
            _unpack_quant_tensor(quant_input),
            _unpack_quant_tensor(quant_weight),
            None,
            *args,
            **kwargs)
    else:
        output = fn(
            _unpack_quant_tensor(quant_input),
            _unpack_quant_tensor(quant_weight),
            _unpack_quant_tensor(bias),
            *args,
            **kwargs)

    if isinstance(quant_input, IntQuantTensor) and isinstance(quant_weight, IntQuantTensor):
        output_bit_width = max_acc_bit_width(
            quant_input.bit_width,
            quant_weight.bit_width,
            quant_weight.value.shape,
            *args,
            **kwargs)
        output_scale = quant_output_scale_impl(
            fn, quant_input.value, quant_input.scale, quant_weight.scale)
        output_signed = quant_input.signed or quant_weight.signed
        output_training = quant_input.training or quant_weight.training

    if bias is not None:
        if output_scale is not None:
            if (isinstance(bias, IntQuantTensor) and
                    not torch.allclose(bias.scale, output_scale)) or not isinstance(bias,
                                                                                    IntQuantTensor):
                channel_dim = -1 if isinstance(fn, torch.nn.Linear) else 1
                output_scale_broadcast_shape = compute_channel_view_shape(
                    quant_input, channel_dim=channel_dim)
                output_zero_point = -_unpack_quant_tensor(bias).view(
                    output_scale_broadcast_shape) / output_scale
        if output_bit_width is not None and isinstance(bias, IntQuantTensor):
            output_bit_width = torch.where(
                bias.bit_width > output_bit_width, bias.bit_width, output_bit_width)
            output_bit_width = output_bit_width + 1

    if compute_output_quant_tensor:
        if (isinstance(quant_input, IntQuantTensor) and
            (quant_input.zero_point != 0.0).any()) or (isinstance(quant_weight, IntQuantTensor) and
                                                       (quant_weight.zero_point != 0.0).any()):
            warnings.warn("Computing zero point of output accumulator not supported yet.")
            compute_output_quant_tensor = False

    if compute_output_quant_tensor:
        if output_zero_point is None:
            output_zero_point = torch.zeros(1).type_as(output)
        return create_int_quant_tensor(
            output,
            output_scale,
            output_bit_width,
            output_zero_point,
            output_signed,
            output_training)
    else:
        return output


def create_int_quant_tensor(tensor, scale, bit_width, zero_point, signed, training):
    from brevitas.quant_tensor import IntQuantTensor
    return IntQuantTensor(
        tensor,
        scale=scale,
        zero_point=zero_point,
        bit_width=bit_width,
        signed=signed,
        training=training)


def quant_output_scale_impl(
        fn: Callable, inp: Tensor, quant_input_scale: Tensor, quant_weight_scale: Tensor):
    channel_dim = -1 if fn == F.linear else 1
    output_scale_shape = compute_channel_view_shape(inp, channel_dim=channel_dim)

    quant_weight_scale = quant_weight_scale.view(output_scale_shape)
    if len(quant_input_scale.shape) == 0:
        quant_input_scale = quant_input_scale.view(output_scale_shape)

    output_scale = quant_weight_scale * quant_input_scale
    return output_scale


def max_acc_bit_width_convNd(input_bit_width, weight_bit_width, weight_shape, *args, **kwargs):
    max_uint_input = max_int(bit_width=input_bit_width, signed=False, narrow_range=False)
    max_kernel_val = max_int(bit_width=weight_bit_width, signed=False, narrow_range=False)
    in_channel = weight_shape[1]
    kernel_size = math.prod(weight_shape[2:])
    max_uint_output = max_uint_input * max_kernel_val * kernel_size * in_channel
    max_output_bit_width = ceil_ste(torch.log2(max_uint_output))
    return max_output_bit_width


def max_acc_bit_width_linear(input_bit_width, weight_bit_width, weight_shape, *args, **kwargs):
    max_uint_input = max_int(bit_width=input_bit_width, signed=False, narrow_range=False)
    max_kernel_val = max_int(bit_width=weight_bit_width, signed=False, narrow_range=False)
    in_channel = weight_shape[1]
    max_uint_output = max_uint_input * max_kernel_val * in_channel
    max_output_bit_width = ceil_ste(torch.log2(max_uint_output))
    return max_output_bit_width


def max_acc_bit_width_convtransposeNd(
        input_bit_width, weight_bit_width, weight_shape, *args, **kwargs):
    stride = kwargs['stride']
    max_uint_input = max_int(bit_width=input_bit_width, signed=False, narrow_range=False)
    max_kernel_val = max_int(bit_width=weight_bit_width, signed=False, narrow_range=False)
    out_channel = weight_shape[1]
    kernel_shape = weight_shape[2:]

    patch_size = 0
    for s, k in zip(stride, kernel_shape):
        patch_size += max(math.ceil(k / s), 1)

    max_uint_output = max_uint_input * max_kernel_val * patch_size * out_channel
    max_output_bit_width = ceil_ste(torch.log2(max_uint_output))
    return max_output_bit_width


def max_acc_bit_width_avg_pool2d(input_bit_width, avg_scaling):
    max_uint_input = max_int(bit_width=input_bit_width, signed=False, narrow_range=False)
    max_uint_output = max_uint_input * avg_scaling
    max_output_bit_width = ceil_ste(torch.log2(max_uint_output))
    return max_output_bit_width


FN_ACC_BITWIDTH_MAPPING = {
    F.linear: max_acc_bit_width_linear,
    F.conv1d: max_acc_bit_width_convNd,
    F.conv2d: max_acc_bit_width_convNd,
    F.conv3d: max_acc_bit_width_convNd,
    F.conv_transpose1d: max_acc_bit_width_convtransposeNd,
    F.conv_transpose2d: max_acc_bit_width_convtransposeNd,
    F.conv_transpose3d: max_acc_bit_width_convtransposeNd,
    F.avg_pool2d: max_acc_bit_width_avg_pool2d}



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from abc import abstractmethod
from copy import copy

import torch
from torch import Tensor

from brevitas.export.common import to_0dim_if_scalar
from brevitas.export.common import to_item_if_0dim
from brevitas.proxy import ActFloatQuantProxyFromInjector
from brevitas.proxy import ActQuantProxyFromInjector
from brevitas.proxy import BiasQuantProxyFromInjector
from brevitas.proxy import DecoupledWeightQuantProxyFromInjector
from brevitas.proxy import DecoupledWeightQuantWithInputProxyFromInjector
from brevitas.proxy import WeightFloatQuantProxyFromInjector
from brevitas.proxy import WeightQuantProxyFromInjector
from brevitas.proxy.float_parameter_quant import BiasFloatQuantProxyFromInjector
from brevitas.proxy.runtime_quant import DynamicActQuantProxyFromInjector
from brevitas.proxy.runtime_quant import TruncQuantProxyFromInjector

from .base import BitWidthHandlerMixin
from .base import ClipMixin
from .base import FloatClipMixin
from .base import FloatZeroPointHandlerMixin
from .base import QuantAxisMixin
from .base import ZeroPointHandlerMixin


class DQMixin(ABC):

    @abstractmethod
    def validate(self):
        pass

    @abstractmethod
    def dequantize_fn(self, x, scale, zero_point, axis):
        pass

    @property
    @abstractmethod
    def flatten_dequantize_params(self):
        pass

    @property
    @abstractmethod
    def itemize_quantize_scalar_params(self):
        pass

    def assert_ge_zero(self, *args):
        for a in args:
            bools = a >= 0.
            if isinstance(bools, torch.Tensor):
                bools = bools.all()
            assert bools


class DQCastMixin(DQMixin, ABC):

    @abstractmethod
    def cast_fn(self, x, dtype):
        pass


class CDQCastMixin(DQCastMixin, ABC):

    @abstractmethod
    def clip_fn(self, x, min_val, max_val):
        pass


class FloatQMixin(ABC):

    @abstractmethod
    def quantize_fn(self, x, scale, zero_point, dtype, axis):
        pass

    @classmethod
    def signed_dtype(cls, exponent_bit_width, mantissa_bit_width, is_ocp, is_fnuz):
        if exponent_bit_width is None or mantissa_bit_width is None:
            return None
        if is_ocp:
            if exponent_bit_width == 4 and mantissa_bit_width == 3:
                dtype = torch.float8_e4m3fn
            elif exponent_bit_width == 5 and mantissa_bit_width == 2:
                dtype = torch.float8_e5m2
        elif is_fnuz:
            if exponent_bit_width == 4 and mantissa_bit_width == 3:
                dtype = torch.float8_e4m3fnuz
            elif exponent_bit_width == 5 and mantissa_bit_width == 2:
                dtype = torch.float8_e5m2fnuz
        else:
            raise NotImplementedError
        return dtype


class QMixin(BitWidthHandlerMixin, ABC):

    @classmethod
    @abstractmethod
    def uint8_dtype(cls):
        pass

    @classmethod
    @abstractmethod
    def int8_dtype(cls):
        pass

    @classmethod
    @abstractmethod
    def int32_dtype(cls):
        pass

    @abstractmethod
    def quantize_fn(self, x, scale, zero_point, dtype, axis):
        pass

    @classmethod
    def signed_dtype(cls, bit_width, is_signed):
        if bit_width is None:
            return None
        if is_signed and bit_width <= 8:
            dtype = cls.int8_dtype()
        elif not is_signed and bit_width <= 8:
            dtype = cls.uint8_dtype()
        elif is_signed and bit_width > 8:
            dtype = cls.int32_dtype()
        else:
            raise RuntimeError(
                "Unsigned quantization > 8b not supported for export, switch to signed.")
        return dtype


class DynamicQMixin(QMixin, ABC):

    @abstractmethod
    def quantize_fn(self, x, dtype):
        pass


class FloatCDQCastProxyHandlerMixin(QuantAxisMixin,
                                    FloatClipMixin,
                                    FloatZeroPointHandlerMixin,
                                    CDQCastMixin,
                                    ABC):

    def dequantize_symbolic_kwargs(
            cls, scale, zero_point, exponent_bit_width, mantissa_bit_width, is_ocp, is_fnuz):
        scale_orig_shape = scale.shape
        axis = cls.quant_axis(scale)
        if cls.flatten_dequantize_params:
            scale = scale.flatten()
        scale = to_0dim_if_scalar(scale)
        if cls.flatten_dequantize_params:
            zero_point = zero_point.flatten()
        zp = to_0dim_if_scalar(zero_point)
        zp = zp.expand_as(scale)
        zp = cls.zero_point_with_dtype(exponent_bit_width, mantissa_bit_width, is_ocp, is_fnuz, zp)
        return {
            'scale': scale,
            'zero_point': zp,
            'axis': axis,
            # We save only the scale original shape
            # as zero-point is being expanded to the same
            # size as the scale
            'scale_orig_shape': scale_orig_shape}


class CDQCastProxyHandlerMixin(QuantAxisMixin, ClipMixin, ZeroPointHandlerMixin, CDQCastMixin, ABC):

    def dequantize_symbolic_kwargs(cls, scale, zero_point, bit_width, is_signed):
        scale_orig_shape = scale.shape
        axis = cls.quant_axis(scale)
        if cls.flatten_dequantize_params:
            scale = scale.flatten()
        scale = to_0dim_if_scalar(scale)
        if cls.flatten_dequantize_params:
            zero_point = zero_point.flatten()
        zp = to_0dim_if_scalar(zero_point)
        zp = zp.expand_as(scale)
        zp = cls.zero_point_with_dtype(is_signed, bit_width, zp)
        return {
            'scale': scale,
            'zero_point': zp,
            'axis': axis,
            # We save only the scale original shape
            # as zero-point is being expanded to the same
            # size as the scale
            'scale_orig_shape': scale_orig_shape}


class FloatQCDQCastWeightQuantProxyHandlerMixin(FloatQMixin, FloatCDQCastProxyHandlerMixin):
    handled_layer = WeightFloatQuantProxyFromInjector

    def quantize_symbolic_kwargs(
            cls, scale, zero_point, exponent_bit_width, mantissa_bit_width, is_ocp, is_fnuz):
        # compute axis before redefining scale
        axis = cls.quant_axis(scale)
        scale = to_0dim_if_scalar(scale.flatten())
        zp = to_0dim_if_scalar(zero_point.flatten())
        # expand_as must go after 0-dim check
        zp = zp.expand_as(scale)
        zp = cls.zero_point_with_dtype(exponent_bit_width, mantissa_bit_width, is_ocp, is_fnuz, zp)
        if cls.itemize_quantize_scalar_params:
            scale = to_item_if_0dim(scale)
            zp = to_item_if_0dim(zp)
        dtype = cls.signed_dtype(exponent_bit_width, mantissa_bit_width, is_ocp, is_fnuz)
        return {'scale': scale, 'zero_point': zp, 'dtype': dtype, 'axis': axis}

    def prepare_quantize_from_floating_point(self, module):
        quant_weight = module.tracked_module_list[0].quant_weight()
        scale = quant_weight.scale
        self.scale_dtype = scale.dtype
        if self.scale_dtype == torch.bfloat16 or self.scale_dtype == torch.float16:
            scale = self.cast_fn(scale, torch.float32)
        self.symbolic_kwargs['quantize_symbolic_kwargs'] = self.quantize_symbolic_kwargs(
            scale,
            quant_weight.zero_point,
            quant_weight.exponent_bit_width,
            quant_weight.mantissa_bit_width,
            module.is_ocp,
            module.is_fnuz)

    def prepare_quantize_from_minifloat(self, module):
        raise NotImplementedError

    def prepare_for_export(self, module):
        if module.is_quant_enabled:
            self.validate(module)
            if self._export_q_node:
                self.prepare_quantize_from_floating_point(module)
            else:
                self.prepare_quantize_from_minifloat(module)
            # Get the first quant weight as representative
            quant_weight = module.tracked_module_list[0].quant_weight()

            # (B)float16 is not supported with standard Q/DQ ops, thus we store the original dtype
            # of the scale and we cast it to float32.
            # The original dtype is then restored during the forward pass
            scale = quant_weight.scale
            self.scale_dtype = scale.dtype
            if self.scale_dtype == torch.bfloat16 or self.scale_dtype == torch.float16:
                scale = self.cast_fn(scale, torch.float32)

            self.symbolic_kwargs['exponent_bit_width'] = quant_weight.exponent_bit_width
            self.symbolic_kwargs['mantissa_bit_width'] = quant_weight.mantissa_bit_width
            self.symbolic_kwargs['exponent_bias'] = quant_weight.exponent_bias
            self.symbolic_kwargs['saturating'] = quant_weight.saturating
            self.symbolic_kwargs['inf_values'] = quant_weight.inf_values
            self.symbolic_kwargs['nan_values'] = quant_weight.nan_values
            self.symbolic_kwargs['clip_symbolic_kwargs'] = self.clip_symbolic_kwargs(
                module.is_narrow_range,
                module.is_signed,
                quant_weight.exponent_bit_width,
                quant_weight.mantissa_bit_width)
            self.symbolic_kwargs['dequantize_symbolic_kwargs'] = self.dequantize_symbolic_kwargs(
                scale,
                quant_weight.zero_point,
                quant_weight.exponent_bit_width,
                quant_weight.mantissa_bit_width,
                module.is_ocp,
                module.is_fnuz)
        else:
            self.symbolic_kwargs = None

    def quantize_from_floating_point(self, x: Tensor):
        # Workaround for equal_cpu RuntimeError
        quantize_symbolic_kwargs = self.symbolic_kwargs['quantize_symbolic_kwargs']
        # Before quantization, cast input to float32
        if self.scale_dtype == torch.float16 or self.scale_dtype == torch.bfloat16:
            x = self.cast_fn(x, torch.float32)
        x = self.quantize_fn(x, *quantize_symbolic_kwargs.values())
        return x

    def quantize_from_minifloat(self, x: Tensor):
        raise NotImplementedError

    def symbolic_execution(self, x: Tensor):
        assert self.symbolic_kwargs is not None, 'Symbolic execution requires quant to be enabled'

        # Copy dict to allow for popping kwargs even on shared quantizers
        dequantize_symbolic_kwargs = copy(self.symbolic_kwargs['dequantize_symbolic_kwargs'])
        scale = dequantize_symbolic_kwargs['scale']
        zero_point = dequantize_symbolic_kwargs['zero_point']

        if self._export_q_node:
            x = self.quantize_from_floating_point(x)
        else:
            x = self.quantize_from_minifloat(x)
        clip_symbolic_kwargs = self.symbolic_kwargs['clip_symbolic_kwargs']
        exponent_bit_width = self.symbolic_kwargs['exponent_bit_width']
        mantissa_bit_width = self.symbolic_kwargs['mantissa_bit_width']
        exponent_bias = self.symbolic_kwargs['exponent_bias']
        saturating = self.symbolic_kwargs['saturating']
        inf_values = self.symbolic_kwargs['inf_values']
        nan_values = self.symbolic_kwargs['nan_values']
        scale_orig_shape = dequantize_symbolic_kwargs.pop('scale_orig_shape')
        # Workaround to trick the tracer into believing all return values are used
        self.assert_ge_zero(scale, exponent_bit_width, mantissa_bit_width, exponent_bias)
        if clip_symbolic_kwargs is not None:
            x = self.clip_fn(x, *clip_symbolic_kwargs.values())
        x = self.dequantize_fn(x, *dequantize_symbolic_kwargs.values())
        # After dequantization, cast both input and scale to the correct dtype
        if self.scale_dtype == torch.float16 or self.scale_dtype == torch.bfloat16:
            x = self.cast_fn(x, self.scale_dtype)
            scale = self.cast_fn(scale, self.scale_dtype)
        # Restore the original shapes to guarantee correct shape propagation downstream
        scale = scale.view(scale_orig_shape)
        zero_point = zero_point.view_as(scale)
        return x, scale, zero_point, exponent_bit_width, mantissa_bit_width, exponent_bias, saturating, inf_values, nan_values


class QCDQCastWeightQuantProxyHandlerMixin(QMixin, CDQCastProxyHandlerMixin):
    handled_layer = WeightQuantProxyFromInjector

    def quantize_symbolic_kwargs(cls, scale, zero_point, bit_width, is_signed):
        # compute axis before redefining scale
        axis = cls.quant_axis(scale)
        scale = to_0dim_if_scalar(scale.flatten())
        zp = to_0dim_if_scalar(zero_point.flatten())
        # expand_as must go after 0-dim check
        zp = zp.expand_as(scale)
        zp = cls.zero_point_with_dtype(is_signed, bit_width, zp)
        if cls.itemize_quantize_scalar_params:
            scale = to_item_if_0dim(scale)
            zp = to_item_if_0dim(zp)
        dtype = cls.signed_dtype(bit_width, is_signed)
        return {'scale': scale, 'zero_point': zp, 'dtype': dtype, 'axis': axis}

    def prepare_quantize_from_floating_point(self, module):
        quant_weight = module.tracked_module_list[0].quant_weight()
        scale = quant_weight.scale
        self.scale_dtype = scale.dtype
        if self.scale_dtype == torch.bfloat16 or self.scale_dtype == torch.float16:
            scale = self.cast_fn(scale, torch.float32)
        self.symbolic_kwargs['quantize_symbolic_kwargs'] = self.quantize_symbolic_kwargs(
            scale, quant_weight.zero_point, quant_weight.bit_width, module.is_signed)

    def prepare_quantize_from_integer(self, module):
        int_weights = {
            tm.weight.data_ptr(): tm.quant_weight().int(float_datatype=False)
            for tm in module.tracked_module_list}
        self.symbolic_kwargs['int_weights'] = int_weights

    def prepare_for_export(self, module):
        if module.is_quant_enabled:
            self.validate(module)
            if self._export_q_node:
                self.prepare_quantize_from_floating_point(module)
            else:
                self.prepare_quantize_from_integer(module)
            # Get the first quant weight as representative
            quant_weight = module.tracked_module_list[0].quant_weight()

            # (B)float16 is not supported with standard Q/DQ ops, thus we store the original dtype
            # of the scale and we cast it to float32.
            # The original dtype is then restored during the forward pass
            scale = quant_weight.scale
            self.scale_dtype = scale.dtype
            if self.scale_dtype == torch.bfloat16 or self.scale_dtype == torch.float16:
                scale = self.cast_fn(scale, torch.float32)

            self.symbolic_kwargs['bit_width'] = quant_weight.bit_width
            self.symbolic_kwargs['clip_symbolic_kwargs'] = self.int_clip_symbolic_kwargs(
                module.is_narrow_range, module.is_signed, quant_weight.bit_width)
            self.symbolic_kwargs['dequantize_symbolic_kwargs'] = self.dequantize_symbolic_kwargs(
                scale, quant_weight.zero_point, quant_weight.bit_width, module.is_signed)
        else:
            self.symbolic_kwargs = None

    def quantize_from_floating_point(self, x: Tensor):
        quantize_symbolic_kwargs = self.symbolic_kwargs['quantize_symbolic_kwargs']
        # Before quantization, cast input to float32
        if self.scale_dtype == torch.float16 or self.scale_dtype == torch.bfloat16:
            x = self.cast_fn(x, torch.float32)
        x = self.quantize_fn(x, *quantize_symbolic_kwargs.values())
        return x

    def quantize_from_integer(self, x: Tensor):
        return self.symbolic_kwargs['int_weights'][x.data_ptr()]

    def symbolic_execution(self, x: Tensor):
        assert self.symbolic_kwargs is not None, 'Symbolic execution requires quant to be enabled'
        if self._export_q_node:
            x = self.quantize_from_floating_point(x)
        else:
            x = self.quantize_from_integer(x)
        clip_symbolic_kwargs = self.symbolic_kwargs['clip_symbolic_kwargs']
        # Copy dict to allow for popping kwargs even on shared quantizers
        dequantize_symbolic_kwargs = copy(self.symbolic_kwargs['dequantize_symbolic_kwargs'])
        scale = dequantize_symbolic_kwargs['scale']
        zero_point = dequantize_symbolic_kwargs['zero_point']
        bit_width = self.symbolic_kwargs['bit_width']
        scale_orig_shape = dequantize_symbolic_kwargs.pop('scale_orig_shape')
        # Workaround to trick the tracer into believing all return values are used
        self.assert_ge_zero(scale, zero_point, bit_width)
        if clip_symbolic_kwargs is not None:
            x = self.clip_fn(x, *clip_symbolic_kwargs.values())
        x = self.dequantize_fn(x, *dequantize_symbolic_kwargs.values())
        # After dequantization, cast both input and scale to the correct dtype
        if self.scale_dtype == torch.float16 or self.scale_dtype == torch.bfloat16:
            x = self.cast_fn(x, self.scale_dtype)
            scale = self.cast_fn(scale, self.scale_dtype)
        # Restore the original shapes to guarantee correct shape propagation downstream
        scale = scale.view(scale_orig_shape)
        zero_point = zero_point.view_as(scale)
        return x, scale, zero_point, bit_width


class QCDQCastDecoupledWeightQuantProxyHandlerMixin(QCDQCastWeightQuantProxyHandlerMixin, ABC):
    handled_layer = DecoupledWeightQuantProxyFromInjector

    def symbolic_execution(self, x: Tensor):
        out, scale, zero_point, bit_width = super().symbolic_execution(x)
        # Return post-rounding scale and zero-point in place of pre-rounding as a placeholder
        # The order of arguments must match the order in the forward method of DecoupledRescalingIntQuant
        return out, scale, zero_point, bit_width, scale, zero_point


class QCDQCastDecoupledWeightQuantWithInputProxyHandlerMixin(
        QCDQCastDecoupledWeightQuantProxyHandlerMixin, ABC):
    handled_layer = DecoupledWeightQuantWithInputProxyFromInjector

    def validate(self, module):
        assert not self._export_q_node, "This proxy requires to export integer weights"

    def symbolic_execution(self, x: Tensor, input_bit_width: torch.Tensor, input_is_signed: bool):
        return super().symbolic_execution(x)


class FloatQCDQCastActQuantProxyHandlerMixin(FloatQMixin, FloatCDQCastProxyHandlerMixin, ABC):
    handled_layer = ActFloatQuantProxyFromInjector

    def quantize_symbolic_kwargs(
            cls, scale, zero_point, exponent_bit_width, mantissa_bit_width, is_ocp, is_fnuz):
        # compute axis before redefining scale
        axis = cls.quant_axis(scale)
        scale = to_0dim_if_scalar(scale.flatten())
        zp = to_0dim_if_scalar(zero_point.flatten())
        # expand_as must go after 0-dim check
        zp = zp.expand_as(scale)
        zp = cls.zero_point_with_dtype(exponent_bit_width, mantissa_bit_width, is_ocp, is_fnuz, zp)
        if cls.itemize_quantize_scalar_params:
            scale = to_item_if_0dim(scale)
            zp = to_item_if_0dim(zp)
        dtype = cls.signed_dtype(exponent_bit_width, mantissa_bit_width, is_ocp, is_fnuz)
        return {'scale': scale, 'zero_point': zp, 'dtype': dtype, 'axis': axis}

    def prepare_for_export(self, module):
        if module.is_quant_enabled:
            self.validate(module)
            self.symbolic_kwargs['exponent_bit_width'] = module.exponent_bit_width()
            self.symbolic_kwargs['mantissa_bit_width'] = module.mantissa_bit_width()
            self.symbolic_kwargs['exponent_bias'] = module.exponent_bias()
            self.symbolic_kwargs['saturating'] = module.is_saturating()
            self.symbolic_kwargs['inf_values'] = module.inf_values()
            self.symbolic_kwargs['nan_values'] = module.nan_values()

            # (B)float16 is not supported with standard Q/DQ ops, thus we store the original dtype
            # of the scale and we cast it to float32.
            # The original dtype is then restored during the forward pass
            scale = module.scale()
            self.scale_dtype = scale.dtype
            if self.scale_dtype == torch.bfloat16 or self.scale_dtype == torch.float16:
                scale = self.cast_fn(scale, torch.float32)

            self.symbolic_kwargs['quantize_symbolic_kwargs'] = self.quantize_symbolic_kwargs(
                scale,
                module.zero_point(),
                module.exponent_bit_width(),
                module.mantissa_bit_width(),
                module.is_ocp,
                module.is_fnuz)
            self.symbolic_kwargs['dequantize_symbolic_kwargs'] = self.dequantize_symbolic_kwargs(
                scale,
                module.zero_point(),
                module.exponent_bit_width(),
                module.mantissa_bit_width(),
                module.is_ocp,
                module.is_fnuz)
            self.symbolic_kwargs['clip_symbolic_kwargs'] = self.clip_symbolic_kwargs(
                module.is_narrow_range,
                module.is_signed,
                module.exponent_bit_width(),
                module.mantissa_bit_width())

        else:
            self.symbolic_kwargs = None

    def symbolic_execution(self, x: Tensor):
        assert self.symbolic_kwargs is not None, 'Symbolic execution requires quant to be enabled'

        # Copy dict to allow for popping kwargs even on shared quantizers
        dequantize_symbolic_kwargs = copy(self.symbolic_kwargs['dequantize_symbolic_kwargs'])
        scale = dequantize_symbolic_kwargs['scale']
        zero_point = dequantize_symbolic_kwargs['zero_point']
        scale_orig_shape = dequantize_symbolic_kwargs.pop('scale_orig_shape')

        quantize_symbolic_kwargs = self.symbolic_kwargs['quantize_symbolic_kwargs']
        clip_symbolic_kwargs = self.symbolic_kwargs['clip_symbolic_kwargs']
        exponent_bit_width = self.symbolic_kwargs['exponent_bit_width']
        mantissa_bit_width = self.symbolic_kwargs['mantissa_bit_width']
        exponent_bias = self.symbolic_kwargs['exponent_bias']
        saturating = self.symbolic_kwargs['saturating']
        inf_values = self.symbolic_kwargs['inf_values']
        nan_values = self.symbolic_kwargs['nan_values']

        self.assert_ge_zero(scale, exponent_bit_width, mantissa_bit_width, exponent_bias)
        # If original dtype of the input is (b)float16, cast the input to float32
        if x.dtype == torch.float16 or x.dtype == torch.bfloat16:
            x = self.cast_fn(x, torch.float32)
        x = self.quantize_fn(x, *quantize_symbolic_kwargs.values())
        if clip_symbolic_kwargs is not None:
            x = self.clip_fn(x, *clip_symbolic_kwargs.values())
        x = self.dequantize_fn(x, *dequantize_symbolic_kwargs.values())
        # After dequantization, cast both output and scale to the correct dtype
        if self.scale_dtype == torch.float16 or self.scale_dtype == torch.bfloat16:
            x = self.cast_fn(x, self.scale_dtype)
            scale = self.cast_fn(scale, self.scale_dtype)
        # Restore the original shapes to guarantee correct shape propagation downstream
        scale = scale.view(scale_orig_shape)
        zero_point = zero_point.view_as(scale)
        return x, scale, zero_point, exponent_bit_width, mantissa_bit_width, exponent_bias, saturating, inf_values, nan_values


class QCDQCastActQuantProxyHandlerMixin(QMixin, CDQCastProxyHandlerMixin, ABC):
    handled_layer = ActQuantProxyFromInjector

    def quantize_symbolic_kwargs(cls, scale, zero_point, bit_width, is_signed):
        # compute axis before redefining scale
        axis = cls.quant_axis(scale)
        scale = to_0dim_if_scalar(scale.flatten())
        zp = to_0dim_if_scalar(zero_point.flatten())
        # expand_as must go after 0-dim check
        zp = zp.expand_as(scale)
        zp = cls.zero_point_with_dtype(is_signed, bit_width, zp)
        if cls.itemize_quantize_scalar_params:
            scale = to_item_if_0dim(scale)
            zp = to_item_if_0dim(zp)
        dtype = cls.signed_dtype(bit_width, is_signed)
        return {'scale': scale, 'zero_point': zp, 'dtype': dtype, 'axis': axis}

    def prepare_for_export(self, module):
        if module.is_quant_enabled:
            self.validate(module)
            self.symbolic_kwargs['bit_width'] = module.bit_width()

            # (B)float16 is not supported with standard Q/DQ ops, thus we store the original dtype
            # of the scale and we cast it to float32.
            # The original dtype is then restored during the forward pass
            scale = module.scale()
            self.scale_dtype = scale.dtype
            if self.scale_dtype == torch.bfloat16 or self.scale_dtype == torch.float16:
                scale = self.cast_fn(scale, torch.float32)

            self.symbolic_kwargs['quantize_symbolic_kwargs'] = self.quantize_symbolic_kwargs(
                scale, module.zero_point(), module.bit_width(), module.is_signed)
            self.symbolic_kwargs['dequantize_symbolic_kwargs'] = self.dequantize_symbolic_kwargs(
                scale, module.zero_point(), module.bit_width(), module.is_signed)
            self.symbolic_kwargs['clip_symbolic_kwargs'] = self.int_clip_symbolic_kwargs(
                module.is_narrow_range, module.is_signed, module.bit_width())

        else:
            self.symbolic_kwargs = None

    def symbolic_execution(self, x: Tensor):
        assert self.symbolic_kwargs is not None, 'Symbolic execution requires quant to be enabled'
        quantize_symbolic_kwargs = self.symbolic_kwargs['quantize_symbolic_kwargs']
        clip_symbolic_kwargs = self.symbolic_kwargs['clip_symbolic_kwargs']
        # Copy dict to allow for popping kwargs even on shared quantizers
        dequantize_symbolic_kwargs = copy(self.symbolic_kwargs['dequantize_symbolic_kwargs'])
        scale = dequantize_symbolic_kwargs['scale']
        zero_point = dequantize_symbolic_kwargs['zero_point']
        scale_orig_shape = dequantize_symbolic_kwargs.pop('scale_orig_shape')
        bit_width = self.symbolic_kwargs['bit_width']
        # Workaround to trick the tracer into believing all return values are used
        self.assert_ge_zero(scale, zero_point, bit_width)
        # If original dtype of the input is (b)float16, cast the input to float32
        if x.dtype == torch.float16 or x.dtype == torch.bfloat16:
            x = self.cast_fn(x, torch.float32)
        x = self.quantize_fn(x, *quantize_symbolic_kwargs.values())
        if clip_symbolic_kwargs is not None:
            x = self.clip_fn(x, *clip_symbolic_kwargs.values())
        x = self.dequantize_fn(x, *dequantize_symbolic_kwargs.values())
        # After dequantization, cast both output and scale to the correct dtype
        if self.scale_dtype == torch.float16 or self.scale_dtype == torch.bfloat16:
            x = self.cast_fn(x, self.scale_dtype)
            scale = self.cast_fn(scale, self.scale_dtype)
        # Restore the original shapes to guarantee correct shape propagation downstream
        scale = scale.view(scale_orig_shape)
        zero_point = zero_point.view_as(scale)
        return x, scale, zero_point, bit_width


class DynamicQDQCastActQuantProxyHandlerMixin(DynamicQMixin, DQCastMixin, ABC):
    handled_layer = DynamicActQuantProxyFromInjector

    def prepare_for_export(self, module):
        if module.is_quant_enabled:
            self.validate(module)
            bit_width = module.bit_width()
            is_signed = module.is_signed
            dtype = self.signed_dtype(bit_width, is_signed)
            self.symbolic_kwargs['bit_width'] = bit_width
            self.symbolic_kwargs['is_signed'] = is_signed
            self.symbolic_kwargs['dtype'] = dtype
        else:
            self.symbolic_kwargs = None

    def symbolic_execution(self, x: Tensor):
        assert self.symbolic_kwargs is not None, 'Symbolic execution requires quant to be enabled'

        bit_width = self.symbolic_kwargs['bit_width']
        int_dtype = self.symbolic_kwargs['dtype']
        # Workaround to trick the tracer into believing all return values are used
        self.assert_ge_zero(bit_width)
        # If original dtype of the input is (b)float16, cast the input to float32
        x_dtype = x.dtype
        if x_dtype == torch.float16 or x_dtype == torch.bfloat16:
            x = self.cast_fn(x, torch.float32)

        x, scale, zero_point = self.quantize_fn(x, int_dtype)

        x = self.dequantize_fn(x, scale, zero_point, None)
        # After dequantization, cast both output and scale to the correct dtype
        if x_dtype == torch.float16 or x_dtype == torch.bfloat16:
            x = self.cast_fn(x, x_dtype)
            scale = self.cast_fn(scale, x_dtype)
        return x, scale, zero_point, bit_width


class FloatCDQCastBiasQuantProxyHandlerMixin(DQCastMixin,
                                             QuantAxisMixin,
                                             FloatZeroPointHandlerMixin,
                                             ABC):
    # TODO: We do not have any bias quantizer so this is not wired to anything.
    # Currently we do not support Minifloat -> DQ export for minifloat.
    # This has to be rewritten to be QDQ
    handled_layer = BiasFloatQuantProxyFromInjector

    def validate(self, module):
        if module.bit_width() is not None:
            assert module.bit_width() > 1., 'Binary quant not supported'
        assert module.is_signed, 'Unsigned bias not supported.'
        assert module.rounding_mode == 'ROUND', 'Only round to nearest even supported.'

    def prepare_for_export(self, module):
        if module.is_quant_enabled:
            self.validate(module)
            int_biases = {
                tm.bias.data_ptr(): tm.quant_bias().minifloat(float_datatype=False)
                for tm in module.tracked_module_list}
            self.symbolic_kwargs = {
                'int_biases': int_biases,
                'scale': module.scale(),
                'zero_point': module.zero_point(),
                'exponent_bit_width': module.exponent_bit_width(),
                'mantissa_bit_width': module.mantissa_bit_width(),
                'exponent_bias': module.exponent_bias(),
                'saturating': module.is_saturating(),
                'inf_values': module.inf_values(),
                'nan_values': module.nan_values()}

        else:
            self.symbolic_kwargs = None

    def symbolic_execution(self, x: Tensor, input_scale=None):
        assert self.symbolic_kwargs is not None, 'Symbolic execution requires quant to be enabled'
        int_bias = self.symbolic_kwargs['int_biases'][x.data_ptr()]
        scale = self.symbolic_kwargs['scale']
        zero_point = self.symbolic_kwargs['zero_point']
        exponent_bit_width = self.symbolic_kwargs['exponent_bit_width']
        mantissa_bit_width = self.symbolic_kwargs['mantissa_bit_width']
        exponent_bias = self.symbolic_kwargs['exponent_bias']
        saturating = self.symbolic_kwargs['saturating']
        inf_values = self.symbolic_kwargs['inf_values']
        nan_values = self.symbolic_kwargs['nan_values']

        assert scale is not None or input_scale is not None, 'Input scale required for bias export'
        if input_scale is not None:
            scale = input_scale
        scale_orig_shape = scale.shape

        quant_axis = self.quant_axis(scale)
        if self.flatten_dequantize_params:
            scale = scale.flatten()
            zero_point = zero_point.flatten()
        scale = to_0dim_if_scalar(scale)
        zero_point = to_0dim_if_scalar(zero_point).expand_as(scale)
        zero_point = self.zero_point_with_dtype(
            True, exponent_bit_width, mantissa_bit_width, zero_point)  # assume signed is True
        # If original dtype of scale is (b)float16, store the original dtype
        # and cast the scale to float32
        scale_dtype = scale.dtype
        if scale_dtype == torch.float16 or scale_dtype == torch.bfloat16:
            scale = self.cast_fn(scale, torch.float32)
        y = self.dequantize_fn(int_bias, scale, zero_point, quant_axis)
        # After dequantization, cast both output and scale to the correct dtype
        if scale_dtype == torch.float16 or scale_dtype == torch.bfloat16:
            y = self.cast_fn(y, scale_dtype)
            scale = self.cast_fn(scale, scale_dtype)
        # Restore the original shapes to guarantee correct shape propagation downstream
        scale = scale.view(scale_orig_shape)
        zero_point = zero_point.view_as(scale)
        return y, scale, zero_point, exponent_bit_width, mantissa_bit_width, exponent_bias, saturating, inf_values, nan_values


class CDQCastBiasQuantProxyHandlerMixin(DQCastMixin, QuantAxisMixin, ZeroPointHandlerMixin, ABC):
    handled_layer = BiasQuantProxyFromInjector

    def validate(self, module):
        if module.bit_width() is not None:
            assert module.bit_width() > 1., 'Binary quant not supported'
        assert module.is_signed, 'Unsigned bias not supported.'
        assert module.rounding_mode == 'ROUND', 'Only round to nearest even supported.'

    def prepare_for_export(self, module):
        if module.is_quant_enabled:
            self.validate(module)
            int_biases = {
                tm.bias.data_ptr(): tm.quant_bias().int(float_datatype=False)
                for tm in module.tracked_module_list}
            self.symbolic_kwargs = {
                'int_biases': int_biases,
                'scale': module.scale(),
                'zero_point': module.zero_point(),
                'bit_width': module.bit_width()}
        else:
            self.symbolic_kwargs = None

    def symbolic_execution(self, x: Tensor, input_scale=None):
        assert self.symbolic_kwargs is not None, 'Symbolic execution requires quant to be enabled'
        int_bias = self.symbolic_kwargs['int_biases'][x.data_ptr()]
        scale = self.symbolic_kwargs['scale']
        bit_width = self.symbolic_kwargs['bit_width']
        zero_point = self.symbolic_kwargs['zero_point']
        assert scale is not None or input_scale is not None, 'Input scale required for bias export'
        if input_scale is not None:
            scale = input_scale
        scale_orig_shape = scale.shape

        quant_axis = self.quant_axis(scale)
        if self.flatten_dequantize_params:
            scale = scale.flatten()
            zero_point = zero_point.flatten()
        scale = to_0dim_if_scalar(scale)
        zero_point = to_0dim_if_scalar(zero_point).expand_as(scale)
        zero_point = self.zero_point_with_dtype(
            True, bit_width, zero_point)  # assume signed is True
        # If original dtype of scale is (b)float16, store the original dtype
        # and cast the scale to float32
        scale_dtype = scale.dtype
        if scale_dtype == torch.float16 or scale_dtype == torch.bfloat16:
            scale = self.cast_fn(scale, torch.float32)
        y = self.dequantize_fn(int_bias, scale, zero_point, quant_axis)
        # After dequantization, cast both output and scale to the correct dtype
        if scale_dtype == torch.float16 or scale_dtype == torch.bfloat16:
            y = self.cast_fn(y, scale_dtype)
            scale = self.cast_fn(scale, scale_dtype)
        # Restore the original shapes to guarantee correct shape propagation downstream
        scale = scale.view(scale_orig_shape)
        zero_point = zero_point.view_as(scale)
        return y, scale, zero_point, bit_width


class QCDQCastTruncQuantProxyHandlerMixin(QuantAxisMixin,
                                          ClipMixin,
                                          ZeroPointHandlerMixin,
                                          QMixin,
                                          CDQCastMixin,
                                          ABC):
    handled_layer = TruncQuantProxyFromInjector

    def prepare_for_export(self, module: TruncQuantProxyFromInjector):
        if module.is_quant_enabled:
            self.validate(module)
            self.symbolic_kwargs = {'output_bit_width': module.bit_width()}

    def symbolic_execution(
            self, x: Tensor, scale: Tensor, zero_point: Tensor, input_bit_width: Tensor,
            signed: Tensor):
        assert self.symbolic_kwargs is not None, 'Symbolic execution requires quant to be enabled'
        output_bit_width = self.symbolic_kwargs['output_bit_width']
        dtype = self.int8_dtype() if signed else self.uint8_dtype()
        trunc_scale = 2.0 ** (input_bit_width - output_bit_width)
        # If original dtype of scale is (b)float16, store the original scale dtype
        # and cast the scale and the input to float32
        scale_dtype = scale.dtype
        if scale_dtype == torch.bfloat16 or scale_dtype == torch.float16:
            scale = self.cast_fn(scale, torch.float32)
        if x.dtype == torch.bfloat16 or x.dtype == torch.float16:
            x = self.cast_fn(x, torch.float32)
        pre_scale = scale * trunc_scale
        flat_pre_scale = to_0dim_if_scalar(pre_scale.flatten())
        flat_scale = to_0dim_if_scalar(scale.flatten())
        zp = to_0dim_if_scalar(zero_point.flatten()).expand_as(flat_scale)
        x = self.quantize_fn(x, flat_pre_scale, zp, dtype, self.quant_axis(pre_scale))
        clip_symbolic_kwargs = self.int_clip_symbolic_kwargs(
            signed=signed, narrow=False, bit_width=output_bit_width)
        if clip_symbolic_kwargs is not None:
            x = self.clip_fn(x, *clip_symbolic_kwargs.values())
        x = self.dequantize_fn(x, flat_scale, zp, self.quant_axis(scale))
        # After dequantization, cast both output and scale to the correct dtype
        if scale_dtype == torch.float16 or scale_dtype == torch.bfloat16:
            x = self.cast_fn(x, scale_dtype)
            scale = self.cast_fn(scale, scale_dtype)
        return x, scale, zero_point, output_bit_width



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from warnings import warn

import torch

from brevitas.export.common.handler.qcdq import CDQCastBiasQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import CDQCastMixin
from brevitas.export.common.handler.qcdq import DQCastMixin
from brevitas.export.common.handler.qcdq import DynamicQDQCastActQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import DynamicQMixin
from brevitas.export.common.handler.qcdq import FloatQCDQCastActQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import FloatQCDQCastWeightQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import FloatQMixin
from brevitas.export.common.handler.qcdq import QCDQCastActQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQCastDecoupledWeightQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import \
    QCDQCastDecoupledWeightQuantWithInputProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQCastTruncQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQCastWeightQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QMixin
from brevitas.export.onnx.handler import ONNXBaseHandler
from brevitas.export.onnx.handler import QuantLSTMLayerHandler
from brevitas.inject.enum import ScalingPerOutputType

from ..function import CastFn
from ..function import DequantizeLinearFn
from ..function import DynamicQuantizeLinearFn
from ..function import IntClipFn
from ..function import QuantizeLinearFn


class StdDQCastONNXMixin(DQCastMixin, ABC):

    def dequantize_fn(self, x, scale, zero_point, axis):
        return DequantizeLinearFn.apply(x, scale, zero_point, axis)

    def cast_fn(self, x, dtype):
        return CastFn.apply(x, dtype)

    @property
    def flatten_dequantize_params(self):
        return True

    @property
    def itemize_quantize_scalar_params(self):
        return False

    def validate(self, module):
        assert module.bit_width() > 1., 'Binary quant not supported'


class StdFloatDQCastONNXMixin(StdDQCastONNXMixin, ABC):

    def validate(self, module):
        assert module.is_ocp or module.is_fnuz, 'Only OCP/FNUZ Standard are supported for FP8 export'


class StdFloatCDQCastONNXMixin(CDQCastMixin, StdFloatDQCastONNXMixin, ABC):

    def clip_fn(self, x, min_val, max_val):
        raise NotImplementedError


class StdCDQCastONNXMixin(CDQCastMixin, StdDQCastONNXMixin, ABC):

    def clip_fn(self, x, min_val, max_val):
        return IntClipFn.apply(x, min_val, max_val)


class StdFloatQCDQCastONNXMixin(FloatQMixin, StdFloatCDQCastONNXMixin, ABC):

    def validate(self, module):
        if getattr(self, '_export_q_node', True):
            assert module.rounding_mode.upper() == 'ROUND', 'Only round to nearest even supported'
        super().validate(module)

    def quantize_fn(self, x, scale, zero_point, dtype, axis):
        return QuantizeLinearFn.apply(x, scale, zero_point, dtype, axis)


class StdQCDQCastONNXMixin(QMixin, StdCDQCastONNXMixin, ABC):

    @classmethod
    def int8_dtype(cls):
        return torch.int8

    @classmethod
    def uint8_dtype(cls):
        return torch.uint8

    @classmethod
    def int32_dtype(cls):
        return torch.int32

    def validate(self, module):
        super().validate(module)
        # ONNX QuantizeLinear supports only 8b output with round to nearest even.
        # Below 8b quantization is supported through clipping.
        if getattr(self, '_export_q_node', True):
            assert module.rounding_mode.upper() == 'ROUND', 'Only round to nearest even supported'
        assert not module.is_groupwise, "Export with Per Group quantization not supported"

        self.validate_8b_bit_width(module.bit_width(), le_then=True)

    def quantize_fn(self, x, scale, zero_point, dtype, axis):
        return QuantizeLinearFn.apply(x, scale, zero_point, dtype, axis)


class StdDynamicQDQCastONNXMixin(DynamicQMixin, StdDQCastONNXMixin, ABC):

    @classmethod
    def int8_dtype(cls):
        return torch.int8

    @classmethod
    def uint8_dtype(cls):
        return torch.uint8

    @classmethod
    def int32_dtype(cls):
        return torch.int32

    def validate(self, module):
        super().validate(module)

        assert module.is_signed == False, "Only unsigned quantization supported"
        assert module.quant_injector.scaling_stats_op == 'min_max', "Only min_max scaling op supported"
        # ONNX QuantizeLinear supports only 8b output with round to nearest even.
        # Below 8b quantization is supported through clipping.
        assert module.rounding_mode.upper() == 'ROUND', 'Only round to nearest even supported'
        # Below 8b quantization is not supported.
        self.validate_8b_bit_width(module.bit_width(), le_then=False)
        # Only per tensor quantization is supported
        assert module.quant_injector.scaling_per_output == ScalingPerOutputType.TENSOR, "Only per tensor scaling supported"

    def quantize_fn(self, x, dtype):
        return DynamicQuantizeLinearFn.apply(x, dtype)


class StdFloatQCDQCastONNXWeightQuantProxyHandler(StdFloatQCDQCastONNXMixin,
                                                  FloatQCDQCastWeightQuantProxyHandlerMixin,
                                                  ONNXBaseHandler):
    _export_q_node = False


class StdQCDQCastONNXWeightQuantProxyHandler(StdQCDQCastONNXMixin,
                                             QCDQCastWeightQuantProxyHandlerMixin,
                                             ONNXBaseHandler):
    _export_q_node = False


class StdQCDQCastONNXDecoupledWeightQuantProxyHandler(StdQCDQCastONNXMixin,
                                                      QCDQCastDecoupledWeightQuantProxyHandlerMixin,
                                                      ONNXBaseHandler):
    _export_q_node = False


class StdQCDQCastONNXDecoupledWeightQuantWithInputProxyHandler(
        StdQCDQCastONNXMixin, QCDQCastDecoupledWeightQuantWithInputProxyHandlerMixin,
        ONNXBaseHandler):
    _export_q_node = False


class StdFloatQCDQCastONNXActQuantProxyHandler(StdFloatQCDQCastONNXMixin,
                                               FloatQCDQCastActQuantProxyHandlerMixin,
                                               ONNXBaseHandler):
    pass


class StdQCDQCastONNXActQuantProxyHandler(StdQCDQCastONNXMixin,
                                          QCDQCastActQuantProxyHandlerMixin,
                                          ONNXBaseHandler):
    pass


class StdDynamicQDQCastONNXActQuantProxyHandler(StdDynamicQDQCastONNXMixin,
                                                DynamicQDQCastActQuantProxyHandlerMixin,
                                                ONNXBaseHandler):
    pass


class StdCDQCastONNXBiasQuantProxyHandler(StdDQCastONNXMixin,
                                          CDQCastBiasQuantProxyHandlerMixin,
                                          ONNXBaseHandler):
    pass


class StdQCDQCastONNXTruncQuantProxyHandler(StdQCDQCastONNXMixin,
                                            QCDQCastTruncQuantProxyHandlerMixin,
                                            ONNXBaseHandler):
    pass


class StdQCDQCastONNXQuantLSTMLayerHandler(QuantLSTMLayerHandler):

    def quantized_cell_symbolic_execution(
            self,
            quant_input,
            quant_hidden_state,
            quant_cell_state,
            quant_weight_ii,
            quant_weight_if,
            quant_weight_ic,
            quant_weight_io,
            quant_weight_hi,
            quant_weight_hf,
            quant_weight_hc,
            quant_weight_ho,
            quant_bias_input,
            quant_bias_forget,
            quant_bias_cell,
            quant_bias_output):
        raise RuntimeError(
            "Quantized LSTM cell is not supported for ONNX QCDQ "
            "(weights only quantization is). Use export_qonnx.")



from brevitas.nn import QuantConv2d
from brevitas.nn import QuantLinear


class LoRACompatibleQuantConv2d(QuantConv2d):
    """
    A QuantConv2d layer that can be used with as a replacement for LoRACompatibleConv.
    It doesn't actually support LoRA, it only matches the same forward pass.
    """

    def forward(self, hidden_states, scale: float = 1.0):
        return super().forward(hidden_states)


class LoRACompatibleQuantLinear(QuantLinear):
    """
    A QuantLinear layer that can be used with as a replacement for LoRACompatibleLinear.
    It doesn't actually support LoRA, it only matches the same forward pass.
    """

    def forward(self, hidden_states, scale: float = 1.0):
        return super().forward(hidden_states)



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

Adapted from https://github.com/mit-han-lab/llm-awq, released under the following LICENSE:

MIT License

Copyright (c) 2023 MIT HAN Lab

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import torch
import torch.nn as nn
from transformers.models.llama.modeling_llama import LlamaDecoderLayer
from transformers.models.llama.modeling_llama import LlamaRMSNorm
from transformers.models.opt.modeling_opt import OPTDecoderLayer

from .utils.module import get_op_by_name
from .utils.module import get_op_name

__all__ = ["auto_scale_block", "apply_scale"]


@torch.no_grad()
def get_weight_scale(weight, q_group_size=-1):
    org_shape = weight.shape
    if q_group_size > 0:
        weight = weight.view(-1, q_group_size)
    scale = weight.abs() / weight.abs().amax(dim=1, keepdim=True)
    scale = scale.view(org_shape)
    scale = scale.mean(0)
    return scale


@torch.no_grad()
def get_act_scale(x):
    return x.abs().view(-1, x.shape[-1]).mean(0)


@torch.no_grad()
def scale_ln_fcs(ln, fcs, scales):
    if not isinstance(fcs, list):
        fcs = [fcs]

    scales = scales.to(ln.weight.device)

    ln.weight.div_(scales)
    if hasattr(ln, 'bias') and ln.bias is not None:
        ln.bias.div_(scales)

    for fc in fcs:
        fc.weight.mul_(scales.view(1, -1))

    for p in ln.parameters():
        assert torch.isnan(p).sum() == 0
    for fc in fcs:
        for p in fc.parameters():
            assert torch.isnan(p).sum() == 0


@torch.no_grad()
def scale_fc_fc(fc1, fc2, scales):
    assert isinstance(fc1, nn.Linear)
    assert isinstance(fc2, nn.Linear)
    assert fc1.out_features == fc2.in_features

    scales = scales.to(fc1.weight.device)

    fc1.weight.div_(scales.view(-1, 1))
    if fc1.bias is not None:
        fc1.bias.div_(scales.view(-1))

    fc2.weight.mul_(scales.view(1, -1))

    for p in fc1.parameters():
        assert torch.isnan(p).sum() == 0
    for p in fc2.parameters():
        assert torch.isnan(p).sum() == 0


@torch.no_grad()
def auto_scale_block(module, module_kwargs, w_bit, q_config, input_feat):
    from .quantizer import pseudo_quantize_tensor

    # firstly, get the weight quantize function
    if w_bit is not None:

        def w_quantize_func(p):
            return pseudo_quantize_tensor(
                p,
                n_bit=w_bit,
                **q_config,
            ).detach()
    else:

        def w_quantize_func(p):
            return p

    if "use_cache" in module_kwargs:
        module_kwargs.pop("use_cache")

    # find the best scale ratio
    def _search_module_scale(block, linears2scale: list, x, kwargs={}):
        # w: co, ci
        # x: n, ci
        x = x.to(next(block.parameters()).device)
        weight = torch.cat([_m.weight for _m in linears2scale], dim=0)
        w_max = get_weight_scale(weight, q_group_size=q_config.get("q_group_size", -1))

        with torch.no_grad():
            org_out = block(x, **kwargs)
            if isinstance(org_out, tuple):
                org_out = org_out[0]

        x_max = get_act_scale(x)

        best_error = float('inf')
        best_ratio = -1
        best_scales = None

        n_grid = 20
        history = []

        org_sd = {k: v.cpu() for k, v in block.state_dict().items()}
        for ratio in range(n_grid):
            ratio = ratio * 1 / n_grid
            scales = (x_max.pow(ratio) / w_max.pow(1 - ratio)).clamp(min=1e-4).view(-1)
            scales = scales / (scales.max() * scales.min()).sqrt()
            for fc in linears2scale:
                fc.weight.mul_(scales.view(1, -1))
                fc.weight.data = w_quantize_func(fc.weight.data) / (scales.view(1, -1))
            out = block(x, **kwargs)
            if isinstance(out, tuple):
                out = out[0]

            loss = (org_out - out).float().pow(2).mean().item()  # float prevents overflow
            history.append(loss)
            is_best = loss < best_error
            if is_best:
                best_error = loss
                best_ratio = ratio
                best_scales = scales
            block.load_state_dict(org_sd)
        if best_ratio == -1:
            print(history)
            raise Exception
        # print(best_ratio)
        best_scales = best_scales.view(-1)

        assert torch.isnan(best_scales).sum() == 0, best_scales
        return best_scales.detach()

    def _auto_get_scale(prev_op, layers, inp, module2inspect=None, kwargs={}):
        # module2inspect: if given, we will check the output diff of this module instead of layers
        if module2inspect is None:
            assert len(layers) == 1
            module2inspect = layers[0]

        scales = _search_module_scale(module2inspect, layers, inp, kwargs)
        # prev_op_name, [layer_name], scale
        return (
            get_op_name(module, prev_op), tuple([get_op_name(module, m) for m in layers]), scales)

    scales_list = []  # return the searched scales

    if isinstance(module, OPTDecoderLayer):
        # attention input
        scales_list.append(
            _auto_get_scale(
                prev_op=module.self_attn_layer_norm,
                layers=[module.self_attn.q_proj, module.self_attn.k_proj, module.self_attn.v_proj],
                inp=input_feat['self_attn.q_proj'],
                module2inspect=module.self_attn,
                kwargs=module_kwargs,
            ))
        # attn out
        scales_list.append(
            _auto_get_scale(
                prev_op=module.self_attn.v_proj,
                layers=[module.self_attn.out_proj],
                inp=input_feat['self_attn.out_proj'],
            ))
        # fc1
        scales_list.append(
            _auto_get_scale(
                prev_op=module.final_layer_norm,
                layers=[module.fc1],
                inp=input_feat['fc1'],
            ))
        # fc2
        scales_list.append(
            _auto_get_scale(
                prev_op=module.fc1,
                layers=[module.fc2],
                inp=input_feat['fc2'],
            ))

    elif isinstance(module, LlamaDecoderLayer):
        # attention input
        scales_list.append(
            _auto_get_scale(
                prev_op=module.input_layernorm,
                layers=[module.self_attn.q_proj, module.self_attn.k_proj, module.self_attn.v_proj],
                inp=input_feat['self_attn.q_proj'],
                module2inspect=module.self_attn,
                kwargs=module_kwargs,
            ))
        # attn out
        scales_list.append(
            _auto_get_scale(
                prev_op=module.self_attn.v_proj,
                layers=[module.self_attn.o_proj],
                inp=input_feat['self_attn.o_proj'],
            ))
        # fc1
        scales_list.append(
            _auto_get_scale(
                prev_op=module.post_attention_layernorm,
                layers=[module.mlp.gate_proj, module.mlp.up_proj],
                inp=input_feat['mlp.gate_proj'],
                module2inspect=module.mlp,
            ))
        # fc2
        scales_list.append(
            _auto_get_scale(
                prev_op=module.mlp.up_proj,
                layers=[module.mlp.down_proj],
                inp=input_feat['mlp.down_proj'],
            ))

    else:
        raise NotImplementedError(f"{type(module)} not supported yet!")

    return scales_list


def apply_scale(module, scales_list, input_feat_dict=None):
    for prev_op_name, layer_names, scales in scales_list:
        prev_op = get_op_by_name(module, prev_op_name)
        layers = [get_op_by_name(module, name) for name in layer_names]

        if isinstance(prev_op, nn.Linear):
            assert len(layers) == 1
            scale_fc_fc(prev_op, layers[0], scales)
        elif isinstance(prev_op, (nn.LayerNorm, LlamaRMSNorm)):
            scale_ln_fcs(prev_op, layers, scales)
        else:
            raise NotImplementedError(f"prev_op {type(prev_op)} not supported yet!")

        # apply the scaling to input feat if given; prepare it for clipping
        if input_feat_dict is not None:
            for layer_name in layer_names:
                inp = input_feat_dict[layer_name]
                inp.div_(scales.view(1, -1).to(inp.device))



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch

from brevitas.quant_tensor import _unpack_quant_tensor
from brevitas.quant_tensor import FloatQuantTensorBase
from brevitas.quant_tensor import QuantTensor
from brevitas.utils.torch_utils import float_internal_scale

from .float_torch_handler import FLOAT_QUANT_TENSOR_FN_HANDLER
from .torch_handler import QUANT_TENSOR_FN_HANDLER

IS_VALID_ATOL = 2e-1
BFLOAT16_IS_VALID_ATOL = 0.5


class FloatQuantTensor(FloatQuantTensorBase, QuantTensor):

    def __new__(
            cls,
            value,
            scale,
            zero_point,
            exponent_bit_width,
            mantissa_bit_width,
            exponent_bias,
            saturating,
            inf_values,
            nan_values,
            signed,
            training):

        if not isinstance(scale, torch.Tensor):
            scale = torch.tensor(scale, dtype=torch.float)
        if not isinstance(zero_point, torch.Tensor):
            zero_point = torch.tensor(zero_point, dtype=torch.float)
        if not isinstance(exponent_bit_width, torch.Tensor):
            exponent_bit_width = torch.tensor(exponent_bit_width, dtype=torch.float)
        if not isinstance(mantissa_bit_width, torch.Tensor):
            mantissa_bit_width = torch.tensor(mantissa_bit_width, dtype=torch.float)
        if not isinstance(exponent_bias, torch.Tensor):
            exponent_bias = torch.tensor(exponent_bias, dtype=torch.float)
        if not isinstance(saturating, torch.Tensor):
            saturating = torch.tensor(saturating, dtype=torch.bool)
        if not isinstance(signed, torch.Tensor):
            signed = torch.tensor(signed, dtype=torch.bool)
        if not isinstance(training, torch.Tensor):
            training = torch.tensor(training, dtype=torch.bool)
        quant_tensor = super().__new__(
            cls,
            value,
            scale,
            zero_point,
            exponent_bit_width,
            mantissa_bit_width,
            exponent_bias,
            saturating,
            inf_values,
            nan_values,
            signed,
            training)
        return quant_tensor

    @property
    def signed(self):
        return self.signed_t.item()

    @property
    def training(self):
        return self.training_t.item()

    @property
    def saturating(self):
        return self.saturating_t.item()

    @property
    def eps(self):
        return torch.finfo(self.scale.dtype).tiny

    def __torch_function__(self, func, types, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        if func in FLOAT_QUANT_TENSOR_FN_HANDLER:
            return FLOAT_QUANT_TENSOR_FN_HANDLER[func](*args, **kwargs)
        elif func in QUANT_TENSOR_FN_HANDLER:
            return QUANT_TENSOR_FN_HANDLER[func](*args, **kwargs)
        else:
            args = _unpack_quant_tensor(args)
            kwargs = _unpack_quant_tensor(kwargs)
            return func(*args, **kwargs)

    @property
    def tensor(self):
        return self.value

    @property
    def _pre_round_float_value(self):
        value = self.value
        scale = self.scale
        if self.scale.dtype == torch.bfloat16:
            value = self.value.type(torch.float32)
            scale = self.scale.type(torch.float32)
        minifloat_value = value / scale
        fp_internal_scale = 1. - self.exponent_bias - self.mantissa_bit_width
        int_scale = float_internal_scale(
            self.value, self.mantissa_bit_width, fp_internal_scale, self.eps)
        minifloat_value = minifloat_value / int_scale
        return minifloat_value

    @property
    def is_valid(self):
        with torch.no_grad():
            pre_round_minifloat_value = self._pre_round_float_value
            rounded_minifloat_value = torch.round(pre_round_minifloat_value)
            max_abs_diff = torch.max(torch.abs(pre_round_minifloat_value - rounded_minifloat_value))
            atol = BFLOAT16_IS_VALID_ATOL if self.value.dtype == torch.bfloat16 else IS_VALID_ATOL
            is_minifloat = max_abs_diff < atol
            # We are missing the checks about self being contained between max and min value
            # given by mantissa, exponent, inf, nan, and saturating
            return is_minifloat

    @property
    def device(self):
        value_device = self.value.device
        is_same_device = True
        for t in [self.scale,
                  self.zero_point,
                  self.exponent_bit_width,
                  self.mantissa_bit_width,
                  self.exponent_bias]:
            is_same_device &= value_device == t.device
        if not is_same_device:
            raise RuntimeError("Value and metadata are on different devices")
        return value_device

    def minifloat(self, float_datatype=True):
        # TODO: Check if OCP and cast to proper data-type if matching
        assert float_datatype, "Minifloat quant returns only higher precision dtype"

        if self.is_valid:
            fp_internal_scale = 1. - self.exponent_bias - self.mantissa_bit_width
            int_scale = float_internal_scale(
                self.value, self.mantissa_bit_width, fp_internal_scale, self.eps)
            float_value = torch.round(self._pre_round_float_value) * int_scale
            return float_value.type(self.scale.dtype)
        else:
            raise RuntimeError(f"FloatQuantTensor not valid.")

    @staticmethod
    def check_input_type(tensor):
        if not isinstance(tensor, FloatQuantTensor):
            raise RuntimeError("Tensor is not a FloatQuantTensor")

    @staticmethod
    def is_zero_zero_point(tensor):
        FloatQuantTensor.check_input_type(tensor)
        return (tensor.zero_point == 0.).all()

    def check_scaling_factors_same(self, other):
        if self.training:
            return True
        if not torch.allclose(self.scale, other.scale):
            raise RuntimeError("Scaling factors are different")

    def check_zero_points_same(self, other):
        if self.training:
            return True
        if not torch.allclose(self.zero_point, other.zero_point):
            raise RuntimeError("Zero points are different")

    def check_bit_width_same(self, other):
        if not torch.allclose(self.exponent_bit_width,
                              other.exponent_bit_width) and not torch.allclose(
                                  self.mantissa_bit_width, other.mantissa_bit_width):
            raise RuntimeError("Bit widths are different")

    def check_exponent_bias(self, other):
        if not torch.allclose(self.exponent_bias, other.exponent_bias):
            raise RuntimeError("Bit widths are different")

    def check_inf_nan_same(self, other):
        if not (set(self.inf_values) == set(other.inf_values)) and not (set(self.nan_values) == set(
                other.nan_values)):
            raise RuntimeError("Floating point representations are different")

    def check_sign_same(self, other):
        if not self.signed == other.signed:
            raise RuntimeError("Signs are different")

    def view(self, *args, **kwargs):
        return self.set(value=self.value.view(*args, **kwargs))

    def reshape(self, *args, **kwargs):
        return self.set(value=self.value.reshape(*args, **kwargs))

    def flatten(self, *args, **kwargs):
        return self.set(value=self.value.flatten(*args, **kwargs))

    def transpose(self, *args, **kwargs):
        value = self.value.transpose(*args, **kwargs)
        tensor_meta = {'scale': self.scale, 'zero_point': self.zero_point}
        for k, tm in tensor_meta.items():
            if len(value.shape) == len(tm.shape):
                tensor_meta[k] = tm.transpose(*args, **kwargs)
        return self.set(value=value, **tensor_meta)

    def permute(self, *args, **kwargs):
        value = self.value.permute(*args, **kwargs)
        tensor_meta = {'scale': self.scale, 'zero_point': self.zero_point}
        for k, tm in tensor_meta.items():
            if len(value.shape) == len(tm.shape):
                tensor_meta[k] = tm.permute(*args, **kwargs)
        return self.set(value=value, **tensor_meta)

    @staticmethod
    def cat(tensors, dim, out=None):
        if out is not None:
            raise RuntimeError("Out not supported.")
        if len(tensors) < 2:
            return tensors[0]
        else:
            first_qt = tensors[0]
            if all([isinstance(qt, FloatQuantTensor) for qt in tensors]):
                for qt in tensors[1:]:
                    first_qt.check_scaling_factors_same(qt)
                    first_qt.check_zero_points_same(qt)
                    first_qt.check_bit_width_same(qt)
                    first_qt.check_exponent_bias(qt)
                    first_qt.check_inf_nan_same(qt)
                    first_qt.check_sign_same(qt)
                output_value = torch.cat([qt.value for qt in tensors], dim=dim)
                output_training = any([qt.training for qt in tensors])
                if output_training:
                    output_scale = sum([qt.scale for qt in tensors]) / len(tensors)
                    output_zero_point = sum([qt.zero_point for qt in tensors]) / len(tensors)
                    output_exponent_bit_width = sum([qt.exponent_bit_width for qt in tensors
                                                    ]) / len(tensors)
                    output_mantissa_bit_width = sum([qt.mantissa_bit_width for qt in tensors
                                                    ]) / len(tensors)
                    output_exponent_bias = sum([qt.exponent_bias for qt in tensors]) / len(tensors)
                else:  # at eval time, they are the same
                    output_scale = first_qt.scale
                    output_zero_point = first_qt.zero_point
                    output_exponent_bit_width = first_qt.exponent_bit_width
                    output_mantissa_bit_width = first_qt.mantissa_bit_width
                    output_exponent_bias = first_qt.exponent_bias
                output_signed = first_qt.signed  # they are the same
                output_saturating = first_qt.saturating  # they are the same
                output_inf_values = first_qt.inf_values  # they are the same
                output_nan_values = first_qt.nan_values  # they are the same
                return FloatQuantTensor(
                    value=output_value,
                    scale=output_scale,
                    zero_point=output_zero_point,
                    exponent_bit_width=output_exponent_bit_width,
                    mantissa_bit_width=output_mantissa_bit_width,
                    exponent_bias=output_exponent_bias,
                    saturating=output_saturating,
                    inf_values=output_inf_values,
                    nan_values=output_nan_values,
                    signed=output_signed,
                    training=output_training)
            else:
                tensors = [_unpack_quant_tensor(qt) for qt in tensors]
                output_value = torch.cat(tensors, dim=dim)
                return output_value

    # Reference: https://docs.python.org/3/reference/datamodel.html#emulating-numeric-types

    def __neg__(self):
        neg_value = (-self.minifloat(float_datatype=True) - self.zero_point) * self.scale
        # In case the dtype of self.minifloat is different from the one of the scale
        neg_value = neg_value.type(self.scale.dtype)
        if self.signed:
            return FloatQuantTensor(
                value=neg_value,
                scale=self.scale,
                zero_point=self.zero_point,
                exponent_bit_width=self.exponent_bit_width,
                mantissa_bit_width=self.mantissa_bit_width,
                exponent_bias=self.exponent_bias,
                saturating=self.saturating,
                inf_values=self.inf_values,
                nan_values=self.nan_values,
                signed=self.signed,
                training=self.training)
        else:
            # TODO: implement
            raise NotImplementedError

    def __add__(self, other):
        if isinstance(other, QuantTensor):
            return self.value + other.value
        else:
            output = self.value + other
        return output

    def __mul__(self, other):
        if isinstance(other, QuantTensor):
            return self.value * other.value
        else:
            output = self.value * other
        return output

    def __str__(self):
        return f"FloatQuantTensor(value={self.value}, scale={self.scale}, zero_point={self.zero_point}, exponent_bit_width={self.exponent_bit_width}, mantissa_bit_width={self.mantissa_bit_width}, exponent_bias={self.exponent_bias}, inf_values={self.inf_values}, nan_values={self.nan_values}, signed_t={self.signed_t}, training_t={self.training_t})"

    def __truediv__(self, other):
        if isinstance(other, QuantTensor):
            return self.value / other.value
        else:
            output = self.value / other
        return output

    def __abs__(self):
        if self.signed:
            abs_value = (
                torch.abs(self.minifloat(float_datatype=True)) - self.zero_point) * self.scale
            # In case the dtype of self.minifloat is different from the one of the scale
            abs_value = abs_value.type(self.scale.dtype)
            return FloatQuantTensor(
                value=abs_value,
                scale=self.scale,
                zero_point=self.zero_point,
                exponent_bit_width=self.exponent_bit_width,
                mantissa_bit_width=self.mantissa_bit_width,
                exponent_bias=self.exponent_bias,
                saturating=self.saturating,
                inf_values=self.inf_values,
                nan_values=self.nan_values,
                signed=False,
                training=self.training)
        else:
            return self



from typing import Any, Tuple

import torch.nn as nn

from brevitas.inject import BaseInjector as Injector
from brevitas.proxy.parameter_quant import WeightQuantProxyFromInjector
from brevitas.quant_tensor import GroupwiseIntQuantTensor
from brevitas.utils.quant_utils import _CachedIOGroupwiseInt


class GroupwiseWeightQuantProxyFromInjector(WeightQuantProxyFromInjector):

    def __init__(self, quant_layer: nn.Module, quant_injector: Injector) -> None:
        super().__init__(quant_layer, quant_injector)
        self.cache_class = _CachedIOGroupwiseInt

    @property
    def group_dim(self):
        return self.quant_injector.group_dim

    @property
    def group_size(self):
        return self.quant_injector.group_size

    def apply_input_view(self, x):
        x = super().apply_input_view(x)
        start_dim = self.group_dim if self.group_dim != -1 else -2
        return x.flatten(start_dim, start_dim + 1)

    def create_quant_tensor(self, qt_args: Tuple[Any]) -> GroupwiseIntQuantTensor:
        out, scale, zero_point, bit_width = qt_args
        return GroupwiseIntQuantTensor(
            out,
            scale,
            zero_point,
            self.group_size,
            self.group_dim,
            bit_width,
            self.is_signed,
            self.training)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import List, Optional, Tuple, Union

import torch
from torch import Tensor
from torch.nn import Module
from torch.nn import Parameter

import brevitas
import brevitas.config as config
from brevitas.core.function_wrapper import Identity
from brevitas.core.function_wrapper import OverBatchOverTensorView
from brevitas.core.restrict_val import _ClampValue
from brevitas.core.restrict_val import _RestrictClampValue
from brevitas.core.restrict_val import _RestrictValue
from brevitas.core.restrict_val import FloatRestrictValue
from brevitas.core.scaling.runtime import _StatsScaling
from brevitas.core.stats import _ParameterListStats
from brevitas.core.stats import _Stats
from brevitas.core.stats import DEFAULT_MOMENTUM
from brevitas.core.stats import SCALAR_SHAPE
from brevitas.core.utils import inplace_momentum_update
from brevitas.core.utils import inplace_tensor_mul
from brevitas.core.utils import StatelessBuffer
from brevitas.function import abs_binary_sign_grad


class ConstScaling(brevitas.jit.ScriptModule):
    """
    ScriptModule implementation of a constant scale factor.

    Args:
        scaling_init (Union[float, Tensor]): value to use as constant scale factor.
        restrict_scaling_impl (Module): restrict scaling_init according to some criteria. Default: None
        scaling_min_val (float): force a lower-bound on scaling_init. Default: None

    Returns:
        Tensor: scale factor wrapped in a float torch.tensor.

    Examples:
        >>> scaling_impl = ConstScaling(1.0)
        >>> scaling_impl(torch.empty(1))
        tensor(1.)
        >>> scaling_impl = ConstScaling(1.0, scaling_min_val=3.0)
        >>> scaling_impl(torch.empty(1))
        tensor(3.)
        >>> scaling_impl = ConstScaling(3.0, restrict_scaling_impl=PowerOfTwoRestrictValue())
        >>> scaling_impl(torch.empty(1))
        tensor(4.)

    Note:
        The forward method accepts a single placeholder argument. This is required by (early versions of)
        TorchScript to be consistent across different scaling implementations.

    Note:
        Maps to scaling_impl_type == ScalingImplType.CONST == 'CONST' == 'const' in higher-level APIs.
    """

    def __init__(
            self,
            scaling_init: Union[float, Tensor],
            restrict_scaling_impl: Module = FloatRestrictValue(),
            scaling_min_val: Optional[float] = None,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None) -> None:
        super(ConstScaling, self).__init__()
        self.restrict_clamp_scaling = _RestrictClampValue(scaling_min_val, restrict_scaling_impl)
        if isinstance(scaling_init, Tensor):
            scaling_init = scaling_init.to(device=device, dtype=dtype)
            scaling_init = restrict_scaling_impl.restrict_init_tensor(scaling_init)
            self.restrict_init_module = restrict_scaling_impl.restrict_init_module()
            self.value = StatelessBuffer(scaling_init.detach())
        else:
            scaling_init = restrict_scaling_impl.restrict_init_float(scaling_init)
            self.restrict_init_module = restrict_scaling_impl.restrict_init_module()
            self.value = StatelessBuffer(torch.tensor(scaling_init, dtype=dtype, device=device))

    @brevitas.jit.script_method
    def forward(self, placeholder: Tensor, threshold: Optional[Tensor] = None) -> Tensor:
        if threshold is None:
            threshold = torch.ones(1).type_as(placeholder)
        # We first apply any restriction to scaling
        # For IntQuant, this is no-op, retrocompatible.
        threshold = self.restrict_clamp_scaling(self.restrict_init_module(threshold))
        restricted_value = self.restrict_clamp_scaling(self.value())
        restricted_value = restricted_value / threshold
        return restricted_value


class ParameterScaling(brevitas.jit.ScriptModule):
    """
    ScriptModule implementation of a learned scale factor.

    Args:
        scaling_init (Union[float, Tensor]): value to initialize the learned scale factor.
        scaling_shape (Tuple[int, ...]): shape to extend a scalar float or tensor scaling_init. Default: None
        restrict_scaling_impl (Module): restrict the learned scale factor according to some criteria. Default: None
        scaling_min_val (float): force a lower-bound on the learned scale factor. Default: None

    Returns:
        Tensor: learned scale factor wrapped in a float torch.tensor.

    Raises:
        RuntimeError: if scaling_init is a non-scalar tensor and scaling_shape is != scaling_init.shape.

    Examples:
        >>> scaling_impl = ParameterScaling(6.0)
        >>> scaling_impl(torch.empty(1))
        tensor(6., grad_fn=<AbsBinarySignGradFnBackward>)
        >>> scaling_impl = ParameterScaling(6.0, scaling_shape=(3,))
        >>> scaling_impl(torch.empty(1))
        tensor([6., 6., 6.], grad_fn=<AbsBinarySignGradFnBackward>)
        >>> scaling_impl = ParameterScaling(6.0, scaling_shape=(3,), restrict_scaling_impl=PowerOfTwoRestrictValue())
        >>> scaling_impl(torch.empty(1))
        tensor([8., 8., 8.], grad_fn=<PowBackward1>)

    Note:
        Set env variable BREVITAS_IGNORE_MISSING_KEYS=1 to avoid errors when retraining
        from a floating point state dict.

    Note:
        The forward method accepts a single placeholder argument. This is required by (early versions of)
        TorchScript to be consistent across different scaling implementations.

    Note:
        Maps to scaling_impl_type == ScalingImplType.PARAMETER == 'PARAMETER' == 'parameter' in higher-level APIs.
    """

    def __init__(
            self,
            scaling_init: Union[float, Tensor],
            scaling_shape: Optional[Tuple[int, ...]] = None,
            restrict_scaling_impl: Module = FloatRestrictValue(),
            scaling_min_val: Optional[float] = None,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None) -> None:
        super(ParameterScaling, self).__init__()

        if (isinstance(scaling_init, Tensor) and scaling_shape is not None and
                scaling_init.shape != SCALAR_SHAPE and scaling_init.shape != scaling_shape):
            raise RuntimeError("scaling_init.shape is non-scalar and != from scaling_shape.")

        if isinstance(scaling_init, Tensor):
            scaling_init = scaling_init.to(device=device, dtype=dtype)
            scaling_init = scaling_init.detach()
        else:
            scaling_init = torch.tensor(scaling_init, dtype=dtype, device=device)

        scaling_init = restrict_scaling_impl.restrict_init_tensor(scaling_init)
        self.restrict_init_module = restrict_scaling_impl.restrict_init_module()

        if scaling_init.shape == SCALAR_SHAPE and scaling_shape is not None:
            scaling_init = torch.full(scaling_shape, scaling_init, dtype=dtype, device=device)
        self.value = Parameter(scaling_init)
        self.restrict_clamp_scaling = _RestrictClampValue(scaling_min_val, restrict_scaling_impl)

    @brevitas.jit.script_method
    def forward(self, placeholder: Tensor, threshold: Optional[Tensor] = None) -> Tensor:
        if threshold is None:
            threshold = torch.ones(1).type_as(placeholder)
        # We first apply any restriction to scaling
        # For IntQuant, this is no-op, retrocompatible.
        threshold = self.restrict_clamp_scaling(self.restrict_init_module(threshold))
        value = abs_binary_sign_grad(self.restrict_clamp_scaling(self.value))
        return value / threshold

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        value_key = prefix + 'value'
        retrocomp_value_key = prefix + 'learned_value'
        if retrocomp_value_key in state_dict:  # retrocompatibility
            state_dict[value_key] = state_dict.pop(retrocomp_value_key)
        super(ParameterScaling, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        if config.IGNORE_MISSING_KEYS and value_key in missing_keys:
            missing_keys.remove(value_key)


class ParameterFromStatsFromParameterScaling(brevitas.jit.ScriptModule):
    """
    ScriptModule implementation of a learned scale factor initialized from statistics of a parameter,
    e.g. weights MSE or AbsMax.
    """

    def __init__(
            self,
            scaling_stats_impl: Module,
            scaling_stats_input_view_shape_impl: Module,
            scaling_stats_input_concat_dim: int,
            tracked_parameter_list: List[torch.nn.Parameter],
            scaling_shape: Tuple[int, ...],
            restrict_scaling_impl: Module = FloatRestrictValue(),
            scaling_min_val: Optional[float] = None,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None) -> None:
        super(ParameterFromStatsFromParameterScaling, self).__init__()
        self.parameter_list_stats = _ParameterListStats(
            scaling_stats_impl,
            scaling_shape,
            scaling_stats_input_view_shape_impl,
            scaling_stats_input_concat_dim,
            tracked_parameter_list)
        self.restrict_scaling_impl = restrict_scaling_impl
        self.stats_scaling_impl = _StatsScaling(
            restrict_scaling_impl, scaling_shape, scaling_min_val, False, False, dtype, device)
        self.init_done: bool = brevitas.jit.Attribute(False, bool)
        self.local_loss_mode: bool = brevitas.jit.Attribute(False, bool)
        self.restrict_inplace_preprocess = restrict_scaling_impl.restrict_init_inplace_module()
        self.restrict_preprocess = restrict_scaling_impl.restrict_init_module()
        self.value = Parameter(torch.full(scaling_shape, 1.0, dtype=dtype, device=device))

    @brevitas.jit.script_method
    def forward(self, ignored: Tensor, threshold: Optional[Tensor] = None) -> Tensor:
        if threshold is None:
            threshold = torch.ones(1).type_as(ignored)
        # Threshold division must happen after we update self.value, but before we apply restrict_preproces
        # This is because we don't want to store a parameter dependant on a runtime value (threshold)
        # And because restrict needs to happen after we divide by threshold
        if self.init_done:
            threshold = self.restrict_inplace_preprocess(threshold)
            value = self.restrict_scaling_impl.combine_scale_threshold(self.value, threshold)
            value = abs_binary_sign_grad(self.stats_scaling_impl.restrict_clamp_scaling(value))
            return value
        else:
            stats = self.parameter_list_stats()
            # workaround to avoid find_ununsed_parameter=True in DDP
            stats = stats + 0. * self.value
            if self.local_loss_mode:
                return self.stats_scaling_impl(stats, threshold)
            stats = self.restrict_inplace_preprocess(stats)
            threshold = self.restrict_inplace_preprocess(threshold)
            inplace_tensor_mul(self.value.detach(), stats)
            value = self.restrict_scaling_impl.combine_scale_threshold(self.value, threshold)
            value = abs_binary_sign_grad(self.stats_scaling_impl.restrict_clamp_scaling(value))
            self.init_done = True
            return value

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        output_dict = super(ParameterFromStatsFromParameterScaling, self).state_dict(
            destination=destination, prefix=prefix, keep_vars=keep_vars)
        # Avoid saving the init value
        if not self.init_done and not config._FULL_STATE_DICT:
            del output_dict[prefix + 'value']
        return output_dict

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        super(ParameterFromStatsFromParameterScaling, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        value_key = prefix + 'value'
        # disable stats collection when a pretrained value is loaded
        if value_key not in missing_keys:
            self.init_done = True
        if config.IGNORE_MISSING_KEYS and value_key in missing_keys:
            missing_keys.remove(value_key)


class ParameterFromRuntimeStatsScaling(brevitas.jit.ScriptModule):
    """
    ScriptModule implementation of a learned scale factor initialized from runtime statistics.
    The implementation works in two phases. During the first phase, statistics are collected in
    the same fashion as batchnorm, meaning that while the module is in training mode a set of per-batch
    statistics are computed and returned, while in background an average of them is retained and returned
    in inference mode. During the second phase, the average accumulated during the first
    phase is used to initialize a learned torch.nn.Parameter, and then the behaviour is the same
    as ParameterScaling.

    Args:
        collect_stats_steps (int): Number of calls to the forward method in training mode to collect statistics for.
        scaling_stats_impl (Module): Implementation of the statistics computed during the collection phase.
        scaling_stats_input_view_shape_impl (Module): Implementation of the view applied to the runtime
            input during the statistics collection phase. Default: OverBatchOverTensorView().
        scaling_shape (Tuple[int, ...]): shape of the torch.nn.Parameter used in the second phase. Default: SCALAR_SHAPE.
        restrict_scaling_impl (Module): restrict the learned scale factor according to some criteria. Default: None
            input before going into scaling_stats_input_view_shape_impl. Default: None
        scaling_stats_momentum: float = Momentum for the statistics moving average. Default: DEFAULT_MOMENTUM.
        scaling_min_val (float): force a lower-bound on the learned scale factor. Default: None.

    Returns:
        Tensor: learned scale factor wrapped in a float torch.tensor.

    Raises:
        RuntimeError: if scaling_shape != SCALAR_SHAPE and scaling_stats_permute_dims is None

    Examples:
        >>> scaling_impl = ParameterFromRuntimeStatsScaling(collect_stats_steps=1, scaling_stats_impl=AbsMax())
        >>> scaling_impl.training
        True
        >>> x = torch.arange(-3, 2, 0.1)
        >>> scaling_impl(x)
        tensor(3.)
        >>> scaling_impl(torch.randn_like(x))
        tensor(3., grad_fn=<AbsBinarySignGradFnBackward>)

    Note:
        Set env variable BREVITAS_IGNORE_MISSING_KEYS=1 to avoid errors when retraining
        from a floating point state dict.

    Note:
        Maps to scaling_impl_type == ScalingImplType.PARAMETER_FROM_STATS == 'PARAMETER_FROM_STATS'
        == 'parameter_from_stats' when applied to runtime values (inputs/outputs/activations) in higher-level APIs.
    """

    def __init__(
            self,
            collect_stats_steps: int,
            scaling_stats_impl: Module,
            scaling_stats_input_view_shape_impl: Module = OverBatchOverTensorView(),
            scaling_shape: Tuple[int, ...] = SCALAR_SHAPE,
            restrict_scaling_impl: Module = FloatRestrictValue(),
            scaling_stats_momentum: Optional[float] = DEFAULT_MOMENTUM,
            scaling_min_val: Optional[float] = None,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None) -> None:
        super(ParameterFromRuntimeStatsScaling, self).__init__()
        assert collect_stats_steps > 0, 'Steps should be more than 0'
        self.collect_stats_steps: int = brevitas.jit.Attribute(collect_stats_steps, int)
        self.counter: int = brevitas.jit.Attribute(0, int)
        self.stats_input_view_shape_impl = scaling_stats_input_view_shape_impl
        self.stats = _Stats(scaling_stats_impl, scaling_shape)
        self.momentum: Optional[float] = brevitas.jit.Attribute(
            scaling_stats_momentum, Optional[float])
        self.register_buffer('buffer', torch.full(scaling_shape, 1.0, dtype=dtype, device=device))
        self.value = Parameter(torch.full(scaling_shape, 1.0, dtype=dtype, device=device))
        self.restrict_scaling_impl = restrict_scaling_impl
        self.restrict_scaling = _RestrictValue(restrict_scaling_impl)
        self.clamp_scaling = _ClampValue(scaling_min_val)
        self.local_loss_mode: bool = brevitas.jit.Attribute(
            False, bool)  # required to support MSE eval or variants
        self.restrict_inplace_preprocess = restrict_scaling_impl.restrict_init_inplace_module()
        self.restrict_preprocess = restrict_scaling_impl.restrict_init_module()

    @brevitas.jit.script_method
    def training_forward(self, stats_input: Tensor, threshold: Tensor) -> Tensor:
        # Threshold division must happen after we update self.value, but before we apply restrict_preproces
        # This is because we don't want to store a parameter dependent on a runtime value (threshold)
        # And because restrict needs to happen after we divide by threshold
        if self.counter < self.collect_stats_steps:
            stats_input = self.stats_input_view_shape_impl(stats_input)
            stats = self.stats(stats_input)
            # workaround to avoid find_ununsed_parameter=True in DDP
            stats = stats + 0. * self.value  # stats gradient will change from None to 0.
            clamped_stats = self.clamp_scaling(stats)
            new_counter = self.counter + 1
            # Whenever we are in local loss mode, we don't update the counter nor the buffer
            if self.local_loss_mode:
                # Local loss mode, we early exit and divide by threshold
                return abs_binary_sign_grad(clamped_stats / threshold)
            if self.counter == 0:
                inplace_tensor_mul(self.buffer, clamped_stats.detach())
            else:
                inplace_momentum_update(
                    self.buffer, clamped_stats.detach(), self.momentum, self.counter, new_counter)
            self.counter = new_counter
            return abs_binary_sign_grad(clamped_stats / threshold)
        elif self.counter == self.collect_stats_steps:
            self.restrict_inplace_preprocess(self.buffer)
            inplace_tensor_mul(self.value.detach(), self.buffer)
            threshold = self.restrict_preprocess(threshold)
            value = self.restrict_scaling_impl.combine_scale_threshold(self.value, threshold)
            self.counter = self.counter + 1
            return abs_binary_sign_grad(self.clamp_scaling(self.restrict_scaling(value)))
        else:
            threshold = self.restrict_preprocess(threshold)
            value = self.restrict_scaling_impl.combine_scale_threshold(self.value, threshold)
            return abs_binary_sign_grad(self.clamp_scaling(self.restrict_scaling(value)))

    @brevitas.jit.script_method
    def forward(self, stats_input: Tensor, threshold: Optional[Tensor] = None) -> Tensor:
        if threshold is None:
            threshold = torch.ones(1).type_as(stats_input)
        if self.training:
            # Threshold division handled inside the training_forward
            return self.training_forward(stats_input, threshold)
        else:
            if self.counter <= self.collect_stats_steps:
                out = self.buffer / threshold
                out = self.restrict_preprocess(out)
            else:
                threshold = self.restrict_preprocess(threshold)
                out = self.restrict_scaling_impl.combine_scale_threshold(self.value, threshold)
            out = abs_binary_sign_grad(self.clamp_scaling(self.restrict_scaling(out)))
        return out

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        output_dict = super(ParameterFromRuntimeStatsScaling, self).state_dict(
            destination=destination, prefix=prefix, keep_vars=keep_vars)
        # Avoid saving the buffer
        del output_dict[prefix + 'buffer']
        # Avoid saving the init value
        if self.counter == 0 and not config._FULL_STATE_DICT:
            del output_dict[prefix + 'value']
        # Save buffer into value for any non-zero number of collection steps
        elif self.counter <= self.collect_stats_steps:
            output_dict[prefix + 'value'] = self.restrict_preprocess(self.buffer)
        return output_dict

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        # Retrocompatibility with older ParameterScaling, for when scaling impl is switched over
        value_key = prefix + 'value'
        retrocomp_value_key = prefix + 'learned_value'
        if retrocomp_value_key in state_dict:
            state_dict[value_key] = state_dict.pop(retrocomp_value_key)

        super(ParameterFromRuntimeStatsScaling, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        # Buffer is supposed to be always missing
        missing_keys.remove(prefix + 'buffer')
        # Pytorch stores training flag as a buffer with JIT enabled
        training_key = prefix + 'training'
        if training_key in missing_keys:
            missing_keys.remove(training_key)
        # disable stats collection when a pretrained value is loaded
        if value_key not in missing_keys:
            self.counter = self.collect_stats_steps + 1
        if config.IGNORE_MISSING_KEYS and value_key in missing_keys:
            missing_keys.remove(value_key)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from .float_parameter_quant import WeightFloatQuantProxyFromInjector
from .float_runtime_quant import ActFloatQuantProxyFromInjector
from .parameter_quant import BiasQuantProxyFromInjector
from .parameter_quant import BiasQuantProxyFromInjectorBase
from .parameter_quant import DecoupledWeightQuantProxyFromInjector
from .parameter_quant import DecoupledWeightQuantWithInputProxyFromInjector
from .parameter_quant import WeightQuantProxyFromInjector
from .parameter_quant import WeightQuantProxyFromInjectorBase
from .runtime_quant import ActQuantProxyFromInjector
from .runtime_quant import ActQuantProxyFromInjectorBase
from .runtime_quant import ClampQuantProxyFromInjector
from .runtime_quant import TruncQuantProxyFromInjector
